from dataclasses import dataclass
from typing import Optional

import lightning.pytorch as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
from sm.misc.funcs import assert_not_null
from torchmetrics import F1Score, Precision, Recall

from gramsplus.misc.dataset import ColumnarDataset
from gramsplus.semanticmodeling.models.pretrained_models import BaseTorchCPAModel
from scripts.www24.stores.cpa_store import CPAStore


@dataclass
class BaseTorchOutput:
    loss: Optional[torch.Tensor]
    prob: torch.Tensor


class EdgeMLPV1(BaseTorchCPAModel):
    VERSION = 100
    EVAL_BATCH_SIZE = 1000
    EXPECTED_ARGS = {"label", "feature", "prop_embed", "header_embed"}
    EXPECTED_EVAL_ARGS = {"feature", "prop_embed", "header_embed"}

    def __init__(
        self,
        base_feat_dim: int,
        embed_dim: int,
    ):
        super().__init__()

        hidden_dim = 3 * base_feat_dim
        self.embed_dim = embed_dim

        self.context_mat_diag = nn.Parameter(torch.ones(self.embed_dim))
        # two hidden layers
        self.cmp = nn.Sequential(
            nn.Linear(base_feat_dim + 1, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, 2),
        )
        self.class_weights = nn.Parameter(
            torch.FloatTensor([1, 1]), requires_grad=False
        )
        self.save_hyperparameters()

        self.metrics = {}
        self.precision = Precision("binary")
        self.recall = Recall("binary")
        self.f1 = F1Score("binary")

    def forward(self, feature, header_embed, prop_embed, label=None) -> BaseTorchOutput:
        if self.embed_dim == 0:
            ctx_score = torch.zeros((feature.shape[0], 1)).to(feature.device)
        else:
            ctx_score = (
                prop_embed * self.context_mat_diag.view(1, -1) * header_embed
            ).sum(dim=1, keepdim=True)

        inputs = torch.cat([feature, ctx_score], dim=1)
        logits = self.cmp(inputs)

        if label is not None:
            loss = F.nll_loss(
                input=F.log_softmax(logits, dim=1),
                target=label,
                weight=self.class_weights,
            )
        else:
            loss = None

        return BaseTorchOutput(loss=loss, prob=F.softmax(logits, dim=1)[:, 1])

    # def make_dataset(self, featstore)
    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-4)
        return optimizer

    def training_step(self, batch, batch_idx):
        output = self._pl_step(batch, batch_idx, "train")
        return output.loss

    def validation_step(self, batch, batch_idx):
        self._pl_step(batch, batch_idx, "val")

    def test_step(self, batch, batch_idx):
        self._pl_step(batch, batch_idx, "test")

    def _pl_step(self, batch, batch_idx, prefix):
        batch_size = batch["label"].shape[0]
        output = self.forward(**{arg: batch[arg] for arg in self.EXPECTED_ARGS})

        self.log(
            f"{prefix}_loss",
            assert_not_null(output.loss),
            prog_bar=True,
            batch_size=batch_size,
        )
        self.precision(output.prob, batch["label"])
        self.recall(output.prob, batch["label"])
        self.f1(output.prob, batch["label"])
        self.log(
            f"{prefix}_precision",
            self.precision,
            batch_size=batch_size,
        )
        self.log(
            f"{prefix}_recall",
            self.recall,
            batch_size=batch_size,
        )
        self.log(
            f"{prefix}_f1",
            self.f1,
            batch_size=batch_size,
        )
        return output



from typing import Dict

from graph.retworkx.api import digraph_all_simple_paths
from sm.dataset import FullTable

from gramsplus.semanticmodeling.postprocessing.cgraph import (
    CGEdgeTriple,
    CGraph,
)
from gramsplus.semanticmodeling.postprocessing.common import (
    add_context,
    ensure_valid_statements,
)
from gramsplus.semanticmodeling.postprocessing.config import PostprocessingConfig


class PairwiseSelection:
    def __init__(
        self,
        table: FullTable,
        cg: CGraph,
        edge_probs: Dict[CGEdgeTriple, float],
        threshold: float,
    ):
        self.table = table
        self.cg = cg
        self.edge_probs = edge_probs
        self.threshold = threshold

    def get_result(self) -> CGraph:
        """Select the highest score relationship between two nodes"""
        edge_probs = {e: p for e, p in self.edge_probs.items() if p >= self.threshold}

        # first step is to remove dangling statements and standalone nodes
        # even if we don't need this step
        subcg = self.cg.subgraph_from_edge_triples(edge_probs.keys())
        subcg.remove_dangling_statement()
        subcg.remove_standalone_nodes()

        compared_nodes = set()
        for u in subcg.nodes():
            if u.is_column_node:
                compared_nodes.add(u.id)
            elif (u.is_entity_node or u.is_literal_node) and u.is_in_context:
                compared_nodes.add(u.id)

        select_edges = set()
        for uid in compared_nodes:
            for vid in compared_nodes:
                paths = [
                    (
                        e1,
                        e2,
                        edge_probs.get((e1.source, e1.target, e1.key), 0.0)
                        + edge_probs.get((e2.source, e2.target, e2.key), 0.0),
                    )
                    for e1, e2 in digraph_all_simple_paths(subcg, uid, vid, cutoff=2)
                ]
                if len(paths) == 0:
                    continue

                e1, e2, score = max(paths, key=lambda x: x[2])
                if score == 0:
                    continue

                select_edges.add(e1.id)
                select_edges.add(e2.id)

        predcg = subcg.subgraph_from_edges(select_edges)
        predcg.remove_dangling_statement()
        predcg.remove_standalone_nodes()

        if PostprocessingConfig.INCLUDE_CONTEXT:
            add_context(subcg, predcg, edge_probs)

        # add back statement property if missing into to ensure a correct model
        ensure_valid_statements(subcg, predcg, create_if_not_exists=True)

        return predcg



from __future__ import annotations

import os
import shutil
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Union

import numpy as np
import orjson
import serde.json
import torch
from hugedict.sqlite import SqliteDict, SqliteDictFieldType
from ream.cache_helper import Cache, MemBackend
from ream.workspace import ReamWorkspace
from sm.misc.funcs import (
    assert_all_item_not_null,
    assert_not_null,
    batch,
    get_incremental_path,
)
from sm.misc.ray_helper import ray_actor_map_1, ray_get_num_gpu
from tqdm.auto import tqdm


class TextEmbedding:
    def __init__(
        self,
        dir: Path,
        index: SqliteDict[str, tuple[int, int]],
        datasets: list[EmbeddingChunk],
        embedding_model: str,
        chunk_size: int,
    ):
        raise Exception("Deprecated")

        self.dir = dir
        # mapping from the text to the index of the dataset and the index of the example
        self.index = index
        self.datasets = datasets

        self.chunk_size = chunk_size
        self.index_buffer = {}
        self.data_buffer = []
        self.embedding_model = embedding_model

        self.available_devices = []
        if torch.cuda.is_available():
            self.n_gpus = torch.cuda.device_count()
            min_mem = EmbeddingModel.get_model_required_memory(self.embedding_model)
            for i in range(torch.cuda.device_count()):
                if torch.cuda.mem_get_info(i)[0] >= min_mem:
                    self.available_devices.append(f"cuda:{i}")
        else:
            self.n_gpus = 0

        if len(self.available_devices) == 0:
            self.available_devices = ["cpu"]

    @classmethod
    def from_disk(
        cls,
        dir: Path,
        embedding_model: str = "sentence-transformers/all-mpnet-base-v2",
        chunk_size: int = 64000,
    ):
        (dir / "datasets").mkdir(parents=True, exist_ok=True)
        index: SqliteDict[str, tuple[int, int]] = SqliteDict(
            dir / "index.sqlite",
            SqliteDictFieldType.str,
            orjson.dumps,
            orjson.loads,
            SqliteDictFieldType.bytes,
        )
        start = 0
        datasets = []
        for dsdir in sorted(
            (dir / "datasets").iterdir(), key=lambda x: int(x.name.split("_")[1])
        ):
            datasets.append(EmbeddingChunk.load(dsdir))
            assert (
                datasets[-1].start == start
            ), f"{dsdir.name} does not start at {start}"
            start = datasets[-1].end

        assert len(index) == sum(len(d) for d in datasets)
        return TextEmbedding(dir, index, datasets, embedding_model, chunk_size)

    def __contains__(self, text: str) -> bool:
        return text in self.index or text in self.index_buffer

    def retrieve(self, text: str) -> Optional[np.ndarray]:
        if text in self.index:
            dsidx, exidx = self.index[text]
            return self.datasets[dsidx][exidx]
        if text in self.index_buffer:
            return self.data_buffer[self.index_buffer[text]]
        return None

    def batch_retrieve_exist(self, texts: list[str]) -> np.ndarray:
        out = []
        for text in texts:
            if text in self.index:
                dsidx, exidx = self.index[text]
                out.append(self.datasets[dsidx][exidx])
            elif text in self.index_buffer:
                out.append(self.data_buffer[self.index_buffer[text]])
            else:
                raise KeyError(text)
        return np.stack(out)

    def batch_get(
        self,
        texts: Union[list[str], np.ndarray, BatchText],
        batch_size: int = 512,
        parallel: bool = True,
        verbose: bool = False,
    ) -> np.ndarray:
        """Get the embeddings for the given texts."""
        if not isinstance(texts, BatchText):
            batch_text = BatchText.from_list_str(texts)
        else:
            batch_text = texts

        all_embs = [
            self.retrieve(text)
            for text in tqdm(
                batch_text.unique_text.keys(),
                desc="retrieving previously computed embeddings",
                disable=not verbose,
            )
        ]
        # we migrate enumerate(batch_text.unique_text.keys()) to batch_text.unique_text.items()
        # unknown_texts = [
        #     (i, text)
        #     for i, text in enumerate(batch_text.unique_text.keys())
        #     if all_embs[i] is None
        # ]
        assert all(
            i == batch_text.unique_text[text]
            for i, text in enumerate(batch_text.unique_text.keys())
        )
        unknown_texts = [
            (i, text)
            for text, i in batch_text.unique_text.items()
            if all_embs[i] is None
        ]

        if len(unknown_texts) > 0:
            # transform those unknown texts into embeddings
            batched_unk_texts = batch(batch_size, unknown_texts)
            num_gpu = int(ray_get_num_gpu())

            # we need to make sure the model has been fetched
            EmbeddingModel.ensure_loaded_model(*self.get_embedding_method_args())

            if parallel and num_gpu > 1 and len(batched_unk_texts) > 1:
                batched_embs: list[np.ndarray] = ray_actor_map_1(
                    EmbeddingModel,
                    "encode_texts",
                    [self.get_embedding_method_args() for _ in range(num_gpu)],
                    [([x[1] for x in b],) for b in batched_unk_texts],
                    verbose=verbose,
                    desc="compute text embeddings",
                    is_actor_remote=False,
                    remote_options={"num_gpus": 1},
                    before_shutdown=np.copy,
                    auto_shutdown=True,
                )
            else:
                emb_method = self.get_embedding_method()
                batched_embs: list[np.ndarray] = [
                    emb_method.encode_texts([x[1] for x in b])
                    for b in tqdm(
                        batched_unk_texts,
                        desc="compute text embeddings",
                        disable=not verbose,
                    )
                ]

            for i in range(len(batched_unk_texts)):
                for (j, unk_text), emb in zip(batched_unk_texts[i], batched_embs[i]):
                    self.index_buffer[unk_text] = len(self.data_buffer)
                    self.data_buffer.append(emb)
                    all_embs[j] = emb

            self.flush(True)

        new_all_embs = assert_all_item_not_null(all_embs)
        return np.stack([new_all_embs[i] for i in batch_text.text_index])

    def flush(self, soft: bool = False):
        """Flush the dataset to disk."""
        if soft and len(self.data_buffer) < self.chunk_size:
            return

        # determine the range in the buffer to save to disk, if hard flush, save everything
        assert len(self.index_buffer) == len(self.data_buffer)
        if len(self.data_buffer) == 0:
            return

        keys = [""] * len(self.index_buffer)
        for text, i in self.index_buffer.items():
            keys[i] = text
        data = self.data_buffer

        # determine if the last chunk is not full, we may need to update it.
        if len(self.datasets) > 0:
            last_chunk_size = len(self.datasets[-1])
            if last_chunk_size < self.chunk_size and (
                not soft or last_chunk_size + len(data) >= self.chunk_size
            ):
                # the last chunk isn't full and we can flush it
                newkeys = keys[: self.chunk_size - last_chunk_size]
                newdata = data[: self.chunk_size - last_chunk_size]

                ds = self.datasets[-1]
                ds_dir = self.dir / f"datasets/chunks_{len(self.datasets):02d}"
                ds_metadata = serde.json.deser(ds_dir / "metadata.json")
                assert (
                    ds_metadata["start"] == ds.start and ds_metadata["end"] == ds.end
                ), "Make sure that we got the right chunk location"
                assert ds.keys is not None, "The last chunk must load keys"

                ds.end += len(newdata)
                ds.keys.extend(newkeys)
                ds.data = np.concatenate([np.load(ds_dir / "data.npy"), newdata])

                ds.save(Path(str(ds_dir) + "_tmp"))
                shutil.rmtree(ds_dir)

                new_items: list[tuple[str, tuple[int, int]]] = []
                for idx, key in enumerate(ds.keys):
                    new_items.append((key, (len(self.datasets) - 1, idx)))
                self.index.batch_insert(new_items)

                os.rename(Path(str(ds_dir) + "_tmp"), ds_dir)

                data = data[self.chunk_size - last_chunk_size :]
                keys = keys[self.chunk_size - last_chunk_size :]

        if soft:
            save_to = int(len(self.data_buffer) / self.chunk_size) * self.chunk_size
        else:
            save_to = len(self.data_buffer)

        self.data_buffer = data[save_to:]
        self.index_buffer = {k: self.index_buffer[k] - save_to for k in keys[save_to:]}

        data = data[:save_to]
        keys = keys[:save_to]
        start = len(self.index)

        # save the data to disk
        for i in tqdm(range(0, len(data), self.chunk_size), desc="saving embeddings"):
            chunk_dir = get_incremental_path(self.dir / f"datasets/chunks")
            chunk_keys = keys[i : i + self.chunk_size]
            chunk_data = np.array(data[i : i + self.chunk_size])

            ds = EmbeddingChunk(
                start,
                start + len(chunk_data),
                keys=chunk_keys,
                data=chunk_data,
            )
            ds.save(chunk_dir)

            new_items: list[tuple[str, tuple[int, int]]] = []
            for idx, key in enumerate(chunk_keys):
                new_items.append((key, (len(self.datasets), idx)))
            self.index.batch_insert(new_items)

            self.datasets.append(ds)
            start += len(chunk_data)

    @Cache.cache(backend=MemBackend())
    def get_embedding_method(self):
        return EmbeddingModel(*self.get_embedding_method_args())

    def get_embedding_method_args(self) -> tuple:
        cache_dir = self.dir / "models"
        cache_dir.mkdir(exist_ok=True, parents=True)
        return self.embedding_model, cache_dir

    def reindex(self, batch_size: int = 512, verbose: bool = False):
        # verify our index
        for key, (dsidx, exidx) in tqdm(self.index.items(), total=len(self.index)):
            dataset = self.datasets[dsidx]
            assert dataset.keys is not None and dataset.keys[exidx] == key
        reindex_res = ray_actor_map_1(
            EmbeddingModel,
            "_reindex",
            [self.get_embedding_method_args() for _ in range(int(ray_get_num_gpu()))],
            [
                (self.dir / f"datasets/chunks_{i:02d}", batch_size)
                for i in range(1, len(self.datasets) + 1)
            ],
            verbose=verbose,
            desc="reindexing",
            is_actor_remote=False,
            remote_options={"num_gpus": 1},
            auto_shutdown=True,
        )
        for i in range(len(self.datasets)):
            if reindex_res[i]:
                print("Reindex chunk", i + 1)


class EmbeddingChunk:
    def __init__(
        self, start: int, end: int, keys: Optional[list[str]], data: np.ndarray
    ):
        self.start = start
        self.end = end
        self.keys = keys
        self.data = data

    def __getitem__(self, idx: int):
        return self.data[idx]

    def __len__(self):
        return self.end - self.start

    def save(self, dir: Path):
        dir.mkdir(parents=True, exist_ok=True)
        assert not (dir / "metadata.json").exists()
        serde.json.ser(
            {
                "start": self.start,
                "end": self.end,
                "shape": list(self.data.shape),
                "dtype": self.data.dtype.name,
            },
            dir / "metadata.json",
        )
        if self.keys is not None:
            serde.json.ser(self.keys, dir / "keys.json.lz4")
        np.save(dir / "data.npy", self.data)

    @staticmethod
    def load(dir: Path, with_keys: bool = True, mem_map: bool = True) -> EmbeddingChunk:
        if with_keys:
            keys = serde.json.deser(dir / "keys.json.lz4")
        else:
            keys = None

        metadata = serde.json.deser(dir / "metadata.json")
        return EmbeddingChunk(
            metadata["start"],
            metadata["end"],
            keys,
            (
                np.lib.format.open_memmap(
                    dir / "data.npy",
                    dtype=np.dtype(metadata["dtype"]),
                    mode="r",
                    shape=tuple(metadata["shape"]),
                )
                if mem_map
                else np.load(dir / "data.npy")
            ),
        )


class EmbeddingModel:
    def __init__(self, embedding_model: str, cache_dir: Path):
        self.model = EmbeddingModel.get_embedding_model(embedding_model, cache_dir)
        self.model_name = embedding_model

    def encode_texts(self, texts: list[str]):
        embs = self.model.encode(texts)
        assert isinstance(embs, np.ndarray)
        return embs

    @staticmethod
    def get_embedding_model(model: str, cache_dir: Path):
        # TODO: it's difficult to specify a particular GPU in ray
        if model.startswith("sentence-transformers/"):
            from sentence_transformers import SentenceTransformer

            return SentenceTransformer(model, cache_folder=str(cache_dir))

        raise NotImplementedError(model)

    @staticmethod
    def ensure_loaded_model(embedding_model: str, cache_dir: Path):
        if (cache_dir / "_DOWNLOADED_SUCCESS").exists():
            return

        EmbeddingModel.get_embedding_model(embedding_model, cache_dir).encode(
            ["hello world"]
        )
        (cache_dir / "_DOWNLOADED_SUCCESS").touch()

    @staticmethod
    def get_model_required_memory(embedding_model: str):
        if embedding_model == "sentence-transformers/all-mpnet-base-v2":
            return 2.5 * (1024**3)  # 2.5GB
        return 1 * (1024**3)  # 1GB

    def _reindex(self, chunk_loc: Path, batch_size: int = 512):
        chunk = EmbeddingChunk.load(chunk_loc, with_keys=True, mem_map=False)
        chunk_emb = np.concatenate(
            [
                self.encode_texts(b)
                for b in batch(batch_size, assert_not_null(chunk.keys))
            ]
        )
        if not np.allclose(chunk_emb, chunk.data, rtol=0.0, atol=1e-8):
            # need to re-index
            if (chunk_loc / "data.npy.bad").exists():
                raise Exception("Previous reindexing does not work")
            np.save(chunk_loc / "data.tmp.npy", chunk_emb)
            os.rename(chunk_loc / "data.npy", chunk_loc / "data.bad.npy")
            os.rename(chunk_loc / "data.tmp.npy", chunk_loc / "data.npy")
            return True
        return False


@dataclass
class BatchText:
    """A utility class for handling a long list of texts with potentially duplicated."""

    unique_text: dict[str, int] = field(default_factory=dict)
    text_index: list[int] = field(default_factory=list)

    @staticmethod
    def from_list_str(texts: list[str] | np.ndarray) -> BatchText:
        batch_text = BatchText()
        for text in texts:
            batch_text.add_text(text)
        return batch_text

    def add_text(self, text: str):
        """Add text to this batch and return its position"""
        if text not in self.unique_text:
            index = len(self.unique_text)
            self.unique_text[text] = index
        else:
            index = self.unique_text[text]

        self.text_index.append(index)
        return index

    def add_repeated_text(self, text: str, n: int):
        if text not in self.unique_text:
            index = len(self.unique_text)
            self.unique_text[text] = index
        else:
            index = self.unique_text[text]

        self.text_index.extend([index] * n)
        return index

    def __len__(self):
        return len(self.text_index)


@dataclass
class BatchTextEmbedding:
    unique_text: dict[str, int]
    text_index: list[int]
    embeddings: np.ndarray

    @staticmethod
    def from_batch_text(texts: list[str] | BatchText, embs: np.ndarray):
        if not isinstance(texts, BatchText):
            batch_text = BatchText.from_list_str(texts)
        else:
            batch_text = texts
        return BatchTextEmbedding(batch_text.unique_text, batch_text.text_index, embs)

    def get_embedding(self, text: str):
        return self.embeddings[self.text_index[self.unique_text[text]]]


_text_embeddings: dict[str, TextEmbedding] = {}


def ream_get_text_embedding(model: str, singleton: bool = True):
    global _text_embeddings

    needinit = not singleton or model not in _text_embeddings
    if needinit:
        fspath = ReamWorkspace.get_instance().fs.get(
            f"embeddings/{model}",
            diskpath=f"embeddings/{model}",
            key={"model": model},
            save_key=True,
        )
        if fspath.exists():
            realdir = fspath.get()
        else:
            with fspath.reserve_and_track() as realdir:
                ...
        emb = TextEmbedding.from_disk(realdir, model)
        if singleton:
            _text_embeddings[model] = emb
    else:
        emb = _text_embeddings[model]

    return emb


@contextmanager
def auto_flush_text_embedding():
    global _text_embeddings

    try:
        yield None
    finally:
        for emb in _text_embeddings.values():
            emb.flush(soft=False)


if __name__ == "__main__":
    from scripts import DATA_DIR

    # validate if the state of the database is correct.
    EMB_DIR = DATA_DIR / "ream/embeddings/sentence_transformers/all_mpnet_base_v2_023"
    emb = TextEmbedding.from_disk(EMB_DIR)
    emb.reindex(verbose=True)

    # for text, (dsidx, exidx) in tqdm(emb.index.items(), desc="text index"):
    #     try:
    #         emb.datasets[dsidx][exidx]
    #     except:
    #         print("Missing", text, dsidx, exidx)
    #         raise



from __future__ import annotations

from typing import Optional, Union

import polars as pl
from gp_core.literal_matchers import LiteralMatcher, LiteralMatcherConfig
from gp_core.models import (
    AlgoContext,
    CGNode,
    GramsDB,
    LocalGramsDB,
    MatchMethod,
    RemoteGramsDB,
    Table,
    TableCells,
)
from kgdata_core.base import RustVecView

def candidate_local_search(
    table: Table,
    context: Optional[AlgoContext],
    cfg: CandidateLocalSearchConfig,
    parallel: bool,
) -> Table: ...
def par_candidate_local_search(
    db: GramsDB,
    tables: list[Table],
    contexts: Optional[list[AlgoContext]],
    cfg: CandidateLocalSearchConfig,
) -> list[Table]: ...
def matching(
    table: Table,
    table_cells: TableCells,
    context: AlgoContext,
    literal_matcher: LiteralMatcher,
    ignored_columns: list[str],
    ignored_props: list[str],
    allow_same_ent_search: bool = False,
    allow_ent_matching: bool = True,
    use_context: bool = True,
    deterministic_order: bool = True,
    parallel: bool = False,
) -> DataMatchesResult: ...
def par_matching(
    db: GramsDB,
    tables: list[Table],
    table_cells: list[TableCells],
    contexts: Optional[list[AlgoContext]],
    literal_matcher: LiteralMatcher,
    ignored_columns: list[list[str]],
    ignored_props: list[str],
    allow_same_ent_search: bool = False,
    allow_ent_matching: bool = True,
    use_context: bool = True,
    deterministic_order: bool = True,
    verbose: bool = False,
) -> list[DataMatchesResult]: ...
def extract_cangraph(
    table: Table,
    cells: TableCells,
    db: GramsDB,
    cfg: CanGraphExtractorConfig,
    context: Optional[AlgoContext] = None,
    parallel: bool = False,
) -> CanGraphExtractedResult: ...
def par_extract_cangraphs(
    tables: list[Table],
    cells: list[TableCells],
    db: GramsDB,
    cfg: CanGraphExtractorConfig,
    context: Optional[list[AlgoContext]] = None,
    verbose: bool = False,
) -> list[CanGraphExtractedResult]: ...
def extract_candidate_entity_link_freqs(
    db: GramsDB,
    table: Table,
    table_cells: TableCells,
    contexts: Optional[AlgoContext],
    literal_matcher: LiteralMatcher,
    ignored_columns: list[str],
    ignored_props: list[str],
    allow_same_ent_search: bool = False,
    allow_ent_matching: bool = True,
    use_context: bool = True,
    deterministic_order: bool = True,
    parallel: bool = False,
) -> list[dict[str, list[int]]]: ...
def par_extract_candidate_entity_link_freqs(
    db: GramsDB,
    tables: list[Table],
    table_cells: list[TableCells],
    contexts: Optional[list[AlgoContext]],
    literal_matcher: LiteralMatcher,
    ignored_columns: list[list[str]],
    ignored_props: list[str],
    allow_same_ent_search: bool = False,
    allow_ent_matching: bool = True,
    use_context: bool = True,
    deterministic_order: bool = True,
    verbose: bool = False,
) -> list[list[dict[str, list[int]]]]: ...

class CandidateLocalSearchConfig:
    strsim: str
    threshold: float
    use_column_name: bool
    use_language: Optional[str]
    search_all_columns: bool

    def __init__(
        self,
        strsim: str,
        threshold: float,
        use_column_name: bool,
        use_language: Optional[str],
        search_all_columns: bool,
    ) -> None: ...

class CanGraphExtractorConfig:
    literal_matcher_config: LiteralMatcherConfig
    ignored_columns: list[int]
    ignored_props: set[str]
    allow_same_ent_search: bool
    allow_ent_matching: bool
    use_context: bool
    add_missing_property: bool
    run_subproperty_inference: bool
    run_transitive_inference: bool
    deterministic_order: bool
    correct_entity_threshold: float
    validate: bool
    n_hop: int

    def __init__(
        self,
        literal_matcher_config: LiteralMatcherConfig,
        ignored_columns: list[int],
        ignored_props: set[str],
        allow_same_ent_search: bool,
        allow_ent_matching: bool,
        use_context: bool,
        add_missing_property: bool,
        run_subproperty_inference: bool,
        run_transitive_inference: bool,
        deterministic_order: bool,
        correct_entity_threshold: float,
        validate: bool,
        n_hop: int,
    ): ...
    def save(self, outfile: str) -> None: ...

class CanGraphExtractedResult:
    nodes: list[CGNode]
    edges: list[str]
    edgedf: pl.DataFrame

class DataMatchesResult:
    def save(self, path: str) -> None: ...
    @staticmethod
    def load(path: str) -> DataMatchesResult: ...
    def get_n_nodes(self) -> int: ...
    def edges(self) -> RustVecView[PotentialRelationships]: ...
    def is_cell_node(self, idx: int) -> bool: ...
    def get_cell_node(self, idx: int) -> CellNode: ...
    def is_entity_node(self, idx: int) -> bool: ...
    def get_entity_node(self, idx: int) -> EntityNode: ...

class CellNode:
    @property
    def col(self) -> int: ...
    @property
    def row(self) -> int: ...

class EntityNode:
    @property
    def entity_id(self) -> str: ...
    @property
    def entity_prob(self) -> float: ...

class PotentialRelationships:
    @property
    def source_id(self) -> int: ...
    @property
    def target_id(self) -> int: ...
    @property
    def rels(self) -> RustVecView[MatchedEntRel]: ...

class MatchedEntRel:
    @property
    def source_entity_id(self) -> str: ...
    @property
    def statements(self) -> RustVecView[MatchedStatement]: ...
    def get_matched_target_entities(self, context: AlgoContext) -> list[str]: ...

PP = property

class MatchedStatement:
    @PP
    def property(self) -> str: ...
    @PP
    def statement_index(self) -> int: ...
    @PP
    def matched_property(self) -> Optional[Match]: ...
    @PP
    def matched_qualifiers(self) -> RustVecView[MatchedQualifier]: ...

class MatchedQualifier:
    @property
    def qualifier(self) -> str: ...
    @property
    def qualifier_index(self) -> int: ...
    @property
    def score(self) -> Match: ...

class Match:
    @property
    def prob(self) -> float: ...
    @property
    def method(self) -> MatchMethod: ...
    def method(self) -> MatchMethod: ...



from gp.entity_linking.candidate_recognition._heuristic_model import HeuristicCanReg
from gp.entity_linking.candidate_recognition._interface import ICanReg
from gp.entity_linking.candidate_recognition._oracle_model import OracleCanReg

__all__ = ["ICanReg", "HeuristicCanReg", "OracleCanReg"]



from __future__ import annotations

import itertools
from copy import copy
from enum import IntEnum
from typing import Literal, Mapping, Optional, Sequence

import numpy as np
from graph.interface import NodeID
from kgdata.models import Ontology
from kgdata.models.ont_property import OntologyProperty
from rdflib import RDFS
from sm.dataset import FullTable
from sm.evaluation import sm_metrics
from sm.namespaces.namespace import KnowledgeGraphNamespace
from sm.outputs.semantic_model import (
    ClassNode,
    DataNode,
    Edge,
    LiteralNode,
    LiteralNodeDataType,
    SemanticModel,
)


class SMNodeType(IntEnum):
    Column = 0
    Class = 1
    Statement = 2
    Entity = 3
    Literal = 4


class SemanticModelHelper:
    DUMMY_CLASS_FOR_INVERSION_URI = "http://wikiba.se/ontology#DummyClassForInversion"
    ID_PROPS = {str(RDFS.label)}

    def __init__(
        self,
        entity_labels: Optional[Mapping[str, str]],
        props: Mapping[str, OntologyProperty],
        kgns: KnowledgeGraphNamespace,
    ):
        self.entity_labels = entity_labels or {}
        self.props = props
        self.kgns = kgns

    @staticmethod
    def from_ontology(ontology: Ontology) -> SemanticModelHelper:
        return SemanticModelHelper(None, ontology.props, ontology.kgns)

    def norm_sm(self, sm: SemanticModel):
        """ "Normalize the semantic model with the following modifications:
        1. Add readable label to edge and class
        2. Convert direct link (without statement) to have statement except the id props.
        """
        new_sm = sm.deep_copy()
        kgns = self.kgns

        # update readable label
        for n in new_sm.iter_nodes():
            if isinstance(n, ClassNode):
                if kgns.has_encrypted_name(n.abs_uri):
                    n.readable_label = self.get_ent_label(kgns.uri_to_id(n.abs_uri))
            elif (
                isinstance(n, LiteralNode) and n.datatype == LiteralNodeDataType.Entity
            ):
                if kgns.has_encrypted_name(n.value):
                    n.readable_label = self.get_ent_label(kgns.uri_to_id(n.value))
        for e in new_sm.iter_edges():
            if e.abs_uri not in self.ID_PROPS and kgns.has_encrypted_name(e.abs_uri):
                e.readable_label = self.get_pnode_label(kgns.uri_to_id(e.abs_uri))

        # convert direct link
        for edge in list(new_sm.iter_edges()):
            if edge.abs_uri in self.ID_PROPS:
                continue
            source = new_sm.get_node(edge.source)
            target = new_sm.get_node(edge.target)

            if (
                not isinstance(source, ClassNode)
                or source.abs_uri != self.kgns.statement_uri
            ) and (
                not isinstance(target, ClassNode)
                or target.abs_uri != self.kgns.statement_uri
            ):
                # this is direct link, we replace its edge
                # assert len(new_sm.get_edges_between_nodes(source.id, target.id)) == 1
                # new_sm.remove_edges_between_nodes(source.id, target.id)

                for edge in new_sm.get_edges_between_nodes(source.id, target.id):
                    new_sm.remove_edge(edge.id)
                    stmt = ClassNode(
                        abs_uri=self.kgns.statement_uri,
                        rel_uri=kgns.get_rel_uri(self.kgns.statement_uri),
                    )
                    new_sm.add_node(stmt)
                    new_sm.add_edge(
                        Edge(
                            source=edge.source,
                            target=stmt.id,
                            abs_uri=edge.abs_uri,
                            rel_uri=edge.rel_uri,
                            approximation=edge.approximation,
                            readable_label=edge.readable_label,
                        )
                    )
                    new_sm.add_edge(
                        Edge(
                            source=stmt.id,
                            target=edge.target,
                            abs_uri=edge.abs_uri,
                            rel_uri=edge.rel_uri,
                            approximation=edge.approximation,
                            readable_label=edge.readable_label,
                        )
                    )
        return new_sm

    def minify_sm(self, sm: SemanticModel):
        """This is a reverse function of `norm_sm`:
        1. Remove an intermediate statement if it doesn't have any qualifiers
        """
        new_sm = sm.copy()

        for n in sm.iter_nodes():
            if isinstance(n, ClassNode) and n.abs_uri == self.kgns.statement_uri:
                inedges = sm.in_edges(n.id)
                outedges = sm.out_edges(n.id)
                if len(outedges) == 1 and outedges[0].abs_uri == inedges[0].abs_uri:
                    # no qualifiers
                    new_sm.remove_node(n.id)
                    for inedge in inedges:
                        assert inedge.abs_uri == outedges[0].abs_uri
                        new_sm.add_edge(
                            Edge(
                                inedge.source,
                                outedges[0].target,
                                inedge.abs_uri,
                                inedge.rel_uri,
                                # just in case user misannotate to not include approximation in both links
                                inedge.approximation or outedges[0].approximation,
                                inedge.readable_label,
                            )
                        )
        return new_sm

    def gen_equivalent_sm(
        self,
        sm: SemanticModel,
        strict: bool = True,
        force_inversion: bool = False,
        limited_invertible_props: Optional[set[str]] = None,
        incorrect_invertible_props: Optional[set[str]] = None,
    ):
        """Given a semantic model (not being modified), generate equivalent **normalized** models by inferring inverse properties.

        Currently, we only inverse the properties, not qualifiers.



        Parameters
        ----------
        sm: the input semantic model (original)
        strict: whether to throw exception when target of an inverse property is not a class.
        force_inversion: only work when strict mode is set to false. Without force_inverse, we skip inverse properties,
                       otherwise, we generate an inverse model with a special class: wikibase:DummyClassForInversion
        limited_invertible_props: if provided, only generate inverse properties for these properties.
        incorrect_invertible_props: if provided, skip generating inverse properties for these properties.
        Returns
        -------
        """
        sm = self.norm_sm(sm)
        kgns = self.kgns

        if incorrect_invertible_props is None:
            incorrect_invertible_props = set()

        invertible_stmts: list[ClassNode] = []
        is_class_fn = lambda n1: isinstance(n1, ClassNode) or (
            isinstance(n1, LiteralNode) and n1.datatype == LiteralNodeDataType.Entity
        )

        for n in sm.iter_nodes():
            if isinstance(n, ClassNode) and n.abs_uri == self.kgns.statement_uri:
                inedges = sm.in_edges(n.id)
                outedges = sm.out_edges(n.id)
                # only has one prop
                (prop,) = list({inedge.abs_uri for inedge in inedges})
                pid = kgns.uri_to_id(prop)
                stmt_has_value = False
                for outedge in outedges:
                    if outedge.abs_uri != prop:
                        # assert len(self.wdprops[self.get_prop_id(outedge.abs_uri)].inverse_properties) == 0, "Just to make sure" \
                        #                                                                                    "that qualifiers is not invertable. Otherwise, this algorithm will missing one generated SMs"
                        # character role has an inverse property: performer. They can be used as qualifier so nothing to do here just pass
                        pass
                    else:
                        stmt_has_value = True
                if (
                    len(self.props[pid].inverse_properties) > 0
                    and pid not in incorrect_invertible_props
                    and (
                        limited_invertible_props is None
                        or pid in limited_invertible_props
                    )
                    and stmt_has_value
                ):
                    # invertible property
                    # people seem to misunderstand what inverse_property means in RDF;
                    # inverse doesn't apply to data property but only object property.
                    # so we catch the error here to detect what we should fix.
                    (outedge,) = [
                        outedge for outedge in outedges if outedge.abs_uri == prop
                    ]
                    targets_are_class = is_class_fn(sm.get_node(outedge.target))
                    if targets_are_class:
                        invertible_stmts.append(n)
                    elif strict:
                        raise Exception(f"{pid} is not invertible")
                    elif force_inversion:
                        assert isinstance(
                            sm.get_node(outedge.target), DataNode
                        ), "Clearly the model is wrong, you have an inverse property to a literal node"
                        invertible_stmts.append(n)

        # we have N statement, so we are going to have N! - 1 ways. It's literally a cartesian product
        all_choices = []
        for stmt in invertible_stmts:
            # assume that each statement only has one incoming link! fix the for loop if this assumption doesn't hold
            (inedge,) = sm.in_edges(stmt.id)
            choice: list[tuple[ClassNode, Optional[str], Optional[str]]] = [
                (stmt, None, None)
            ]
            for invprop in self.props[
                kgns.uri_to_id(inedge.abs_uri)
            ].inverse_properties:
                choice.append(
                    (
                        stmt,
                        (tmp_abs_uri := kgns.id_to_uri(invprop)),
                        kgns.get_rel_uri(tmp_abs_uri),
                    )
                )
            all_choices.append(choice)

        n_choices = np.prod([len(c) for c in all_choices]) - 1
        if n_choices > 256:
            raise sm_metrics.PermutationExplosion("Too many possible semantic models")

        all_choices_perm: list[
            tuple[tuple[ClassNode, Optional[str], Optional[str]]]
        ] = list(
            itertools.product(*all_choices)
        )  # type: ignore
        assert all(
            invprop is None for _, invprop, _ in all_choices_perm[0]
        ), "First choice is always the current semantic model"
        new_sms = [sm]
        for choice_perm in all_choices_perm[1:]:
            new_sm = sm.copy()
            # we now change the statement from original prop to use the inverse prop (change direction)
            # if the invprop is not None
            for stmt, invprop_abs_uri, invprop_rel_uri in choice_perm:
                if invprop_abs_uri is None or invprop_rel_uri is None:
                    continue
                readable_label = self.get_pnode_label(kgns.uri_to_id(invprop_abs_uri))
                # assume that each statement only has one incoming link! fix the for loop if this assumption doesn't hold
                (inedge,) = sm.in_edges(stmt.id)
                # statement must have only one property
                (outedge,) = [
                    outedge
                    for outedge in sm.out_edges(stmt.id)
                    if outedge.abs_uri == inedge.abs_uri
                ]
                assert (
                    len(new_sm.get_edges_between_nodes(inedge.source, stmt.id)) == 1
                    and len(new_sm.get_edges_between_nodes(stmt.id, outedge.target))
                    == 1
                )
                new_sm.remove_edges_between_nodes(inedge.source, stmt.id)
                new_sm.remove_edges_between_nodes(stmt.id, outedge.target)

                target = sm.get_node(outedge.target)
                if not is_class_fn(target):
                    assert isinstance(target, DataNode)
                    dummy_class_node = ClassNode(
                        abs_uri=self.DUMMY_CLASS_FOR_INVERSION_URI,
                        rel_uri=kgns.get_rel_uri(self.DUMMY_CLASS_FOR_INVERSION_URI),
                    )
                    new_sm.add_node(dummy_class_node)
                    new_sm.add_edge(
                        Edge(
                            source=dummy_class_node.id,
                            target=target.id,
                            abs_uri=str(RDFS.label),
                            rel_uri="rdfs:label",
                        )
                    )
                    outedge_target = dummy_class_node.id
                else:
                    outedge_target = outedge.target
                new_sm.add_edge(
                    Edge(
                        source=outedge_target,
                        target=stmt.id,
                        abs_uri=invprop_abs_uri,
                        rel_uri=invprop_rel_uri,
                        approximation=outedge.approximation,
                        readable_label=readable_label,
                    )
                )
                new_sm.add_edge(
                    Edge(
                        source=stmt.id,
                        target=inedge.source,
                        abs_uri=invprop_abs_uri,
                        rel_uri=invprop_rel_uri,
                        approximation=inedge.approximation,
                        readable_label=readable_label,
                    )
                )
            new_sms.append(new_sm)
        return new_sms

    def get_entity_columns(self, sm: SemanticModel) -> list[int]:
        ent_columns = []
        for dnode in sm.iter_nodes():
            if isinstance(dnode, DataNode):
                inedges = sm.in_edges(dnode.id)
                if len(inedges) == 0:
                    continue
                assert len({edge.abs_uri for edge in inedges}) == 1, inedges
                edge_abs_uri = inedges[0].abs_uri
                if edge_abs_uri in self.ID_PROPS:
                    assert len(inedges) == 1, inedges
                    source = sm.get_node(inedges[0].source)
                    assert (
                        isinstance(source, ClassNode)
                        and not source.abs_uri == self.kgns.statement_uri
                    )
                    ent_columns.append(dnode.col_index)
        return ent_columns

    @classmethod
    def is_uri_column(cls, uri: str):
        """Test if an uri is for specifying the column"""
        return uri.startswith("http://example.com/table/")

    @staticmethod
    def get_column_uri(column_index: int):
        return f"http://example.com/table/{column_index}"

    @staticmethod
    def get_column_index(uri: str):
        assert SemanticModelHelper.is_uri_column(uri)
        return int(uri.replace("http://example.com/table/", ""))

    def extract_claims(
        self, tbl: FullTable, sm: SemanticModel, allow_multiple_ent: bool = True
    ):
        """Extract claims from the table given a semantic model.

        If an entity doesn't have link, its id will be null
        """
        # norm the semantic model first
        sm = self.norm_sm(sm)
        kgns = self.kgns
        schemas = {}
        for u in sm.iter_nodes():
            if not isinstance(u, ClassNode) or kgns.statement_uri == u.abs_uri:
                continue

            schema = {"props": {}, "subject": None, "sm_node_id": u.id}
            for us_edge in sm.out_edges(u.id):
                if us_edge.abs_uri in self.ID_PROPS:
                    v = sm.get_node(us_edge.target)
                    assert isinstance(v, DataNode)
                    assert schema["subject"] is None
                    schema["subject"] = v.col_index
                    continue

                s = sm.get_node(us_edge.target)
                assert isinstance(s, ClassNode) and kgns.statement_uri == s.abs_uri
                assert kgns.is_uri_in_main_ns(us_edge.abs_uri)

                pnode = kgns.uri_to_id(us_edge.abs_uri)
                if pnode not in schema["props"]:
                    schema["props"][pnode] = []

                stmt = {
                    "index": len(schema["props"][pnode]),
                    "value": None,
                    "qualifiers": [],
                }
                schema["props"][pnode].append(stmt)
                for sv_edge in sm.out_edges(s.id):
                    v = sm.get_node(sv_edge.target)

                    assert kgns.is_uri_in_main_ns(sv_edge.abs_uri)
                    if sv_edge.abs_uri == us_edge.abs_uri:
                        assert stmt["value"] is None, "only one property"
                        # this is property
                        if isinstance(v, ClassNode):
                            stmt["value"] = {"type": "classnode", "value": v.id}
                        elif isinstance(v, DataNode):
                            stmt["value"] = {"type": "datanode", "value": v.col_index}
                        else:
                            assert isinstance(v, LiteralNode)
                            stmt["value"] = {"type": "literalnode", "value": v.value}
                    else:
                        # this is qualifier
                        if isinstance(v, ClassNode):
                            stmt["qualifiers"].append(
                                {
                                    "type": "classnode",
                                    "pnode": kgns.uri_to_id(sv_edge.abs_uri),
                                    "value": v.id,
                                }
                            )
                        elif isinstance(v, DataNode):
                            stmt["qualifiers"].append(
                                {
                                    "type": "datanode",
                                    "pnode": kgns.uri_to_id(sv_edge.abs_uri),
                                    "value": v.col_index,
                                }
                            )
                        else:
                            assert isinstance(v, LiteralNode)
                            stmt["qualifiers"].append(
                                {
                                    "type": "literalnode",
                                    "pnode": kgns.uri_to_id(sv_edge.abs_uri),
                                    "value": v.value,
                                }
                            )
            schemas[u.id] = schema

        assert all(
            c.index == ci for ci, c in enumerate(tbl.table.columns)
        ), "Cannot handle table with missing columns yet"

        records = [{} for _ in range(tbl.table.nrows())]
        node2ents = {}

        # extract data props first
        for cid, schema in schemas.items():
            ci: int = schema["subject"]
            col = tbl.table.get_column_by_index(ci)
            for ri, val in enumerate(col.values):
                # get entities
                qnode_ids = sorted(
                    {
                        entity_id
                        for link in tbl.links[ri][ci]
                        for entity_id in link.entities
                        if link.start < link.end
                    }
                )
                if len(qnode_ids) == 0:
                    # create new entity
                    ents = [
                        {
                            "id": f"{ri}-{ci}",
                            "column": ci,
                            "row": ri,
                            "uri": None,
                            "label": val,
                            "props": {},
                        }
                    ]
                else:
                    ents = [
                        {
                            "id": qnode_id,
                            "uri": kgns.id_to_uri(qnode_id),
                            "label": self.get_ent_label(qnode_id),
                            "props": {},
                        }
                        for qnode_id in qnode_ids
                    ]

                if len(ents) > 1:
                    if not allow_multiple_ent:
                        raise Exception("Encounter multiple entities")

                for prop, stmts in schema["props"].items():
                    for ent in ents:
                        assert prop not in ent["props"]
                        ent["props"][prop] = [
                            {"value": None, "qualifiers": {}} for stmt in stmts
                        ]
                        for stmt in stmts:
                            # set statement property
                            if stmt["value"]["type"] == "classnode":
                                # do it in later phase
                                pass
                            elif stmt["value"]["type"] == "datanode":
                                tci = stmt["value"]["value"]
                                ent["props"][prop][stmt["index"]]["value"] = (
                                    tbl.table.get_column_by_index(tci).values[ri]
                                )
                            else:
                                assert stmt["value"]["type"] == "literalnode"
                                ent["props"][prop][stmt["index"]]["value"] = stmt[
                                    "value"
                                ]["value"]

                            # set statement qualifiers
                            for qual in stmt["qualifiers"]:
                                if (
                                    qual["pnode"]
                                    not in ent["props"][prop][stmt["index"]][
                                        "qualifiers"
                                    ]
                                ):
                                    ent["props"][prop][stmt["index"]]["qualifiers"][
                                        qual["pnode"]
                                    ] = []
                                if qual["type"] == "classnode":
                                    # do it in later phase
                                    pass
                                elif qual["type"] == "datanode":
                                    tci = qual["value"]
                                    ent["props"][prop][stmt["index"]]["qualifiers"][
                                        qual["pnode"]
                                    ].append(
                                        tbl.table.get_column_by_index(tci).values[ri]
                                    )
                                elif qual["type"] == "literalnode":
                                    ent["props"][prop][stmt["index"]]["qualifiers"][
                                        qual["pnode"]
                                    ].append(qual["value"])

                for ent in ents:
                    assert (ent["id"], ci) not in records[ri]
                    records[ri][ent["id"], ci] = ent
                node2ents[schema["sm_node_id"], ri] = [ent for ent in ents]

        for cid, schema in schemas.items():
            ci = schema["subject"]
            col = tbl.table.get_column_by_index(ci)
            for ri in range(len(col.values)):
                ulst = node2ents[schema["sm_node_id"], ri]
                for prop, stmts in schema["props"].items():
                    for stmt in stmts:
                        if stmt["value"]["type"] == "classnode":
                            vlst = node2ents[stmt["value"]["value"], ri]
                            for u in ulst:
                                assert len(vlst) > 0
                                u["props"][prop][stmt["index"]]["value"] = vlst[0]["id"]
                                if len(vlst) > 1:
                                    # this statement must not have other qualifiers, so that v can be a list
                                    # and we can create extra statement
                                    assert len(stmt["qualifiers"]) == 0
                                    for v in vlst[1:]:
                                        u["props"][prop].append(
                                            {"value": v["id"], "qualifiers": {}}
                                        )
                        for qual in stmt["qualifiers"]:
                            if qual["type"] == "classnode":
                                for u in ulst:
                                    vlst = node2ents[qual["value"], ri]
                                    u["props"][prop][stmt["index"]]["qualifiers"][
                                        qual["pnode"]
                                    ] = [v["id"] for v in vlst]

        new_records = []
        for ri, record in enumerate(records):
            new_record = {}
            for (ent_id, ci), ent in record.items():
                if ent_id in new_record:
                    # merge the entity
                    for pid, _stmts in ent["props"].items():
                        if pid not in new_record[ent_id]["props"]:
                            new_record[ent_id]["props"][pid] = _stmts
                        else:
                            for _stmt in _stmts:
                                if not any(
                                    x == _stmt for x in new_record[ent_id]["props"][pid]
                                ):
                                    new_record[ent_id]["props"][pid].append(_stmt)
                else:
                    new_record[ent_id] = ent
            new_records.append(new_record)
        return records

    def get_ent_label(self, eid: str):
        """Get WDEntity label from id"""
        if eid not in self.entity_labels:
            return eid
        return f"{self.entity_labels[eid]} ({eid})"

    def get_pnode_label(self, pid: str):
        """Get PNode label from id"""
        if pid not in self.props:
            return pid
        return f"{self.props[pid].label} ({pid})"

    def create_sm(
        self,
        nodes: Mapping[NodeID, ClassNode | DataNode | LiteralNode],
        cpa: Sequence[tuple[NodeID, NodeID, str]],
        cta: Mapping[int, str],
        validate: bool = True,
        on_untype_source_column_node: Literal[
            "create-class", "remove-link"
        ] = "create-class",
    ) -> SemanticModel:
        """Create a semantic model from outputs of CPA and CTA tasks

        # Arguments
            nodes: mapping from node id to DataNode or LiteralNode (leaf nodes) or Statement (represent n-ary relation)
                that we can't generate automatically from the cta. Sources and targets in CPA should all be in this map.
            cpa: list of triples (source, target, predicate)
            cta: mapping from column index to class id
            validate: whether to validate the input to make sure it's correct classes and properties
        """
        if validate:
            for source, target, predicate in cpa:
                assert self.kgns.is_id(predicate)
            for col_index, ent_id in cta.items():
                assert self.kgns.is_id(ent_id)
            for node in nodes.values():
                if isinstance(node, ClassNode):
                    assert self.kgns.is_uri_in_ns(node.abs_uri)
                elif (
                    isinstance(node, LiteralNode)
                    and node.datatype == LiteralNodeDataType.Entity
                ):
                    assert self.kgns.is_uri(node.value) and self.kgns.is_uri_in_main_ns(
                        node.value
                    )

        sm = SemanticModel()
        # make a copy because when we add node into sm, we will modify the node id
        nodes = {uid: copy(u) for uid, u in nodes.items()}

        classmap = {}  # mapping from column to its class node
        # mapping from node id to sm node id, this mapping is built dynamically when iterating over cpa result.
        nodemap: dict[NodeID, int] = {}
        kgns = self.kgns

        col2id = {
            u.col_index: uid for uid, u in nodes.items() if isinstance(u, DataNode)
        }
        for col_index, ent_id in cta.items():

            # somehow, they may end-up predict multiple classes, we need to select one
            if ent_id.find(" ") != -1:
                ent_id = ent_id.split(" ")[0]
            curl = kgns.id_to_uri(ent_id)

            try:
                cnode_label = self.get_ent_label(ent_id)
            except KeyError:
                cnode_label = kgns.get_rel_uri(kgns.id_to_uri(ent_id))

            cnode = ClassNode(
                abs_uri=curl,
                rel_uri=kgns.get_rel_uri(kgns.id_to_uri(ent_id)),
                readable_label=cnode_label,
            )
            dnode = nodes[col2id[col_index]]
            sm.add_node(cnode)
            nodemap[col2id[col_index]] = sm.add_node(dnode)
            classmap[col_index] = cnode.id
            sm.add_edge(
                Edge(
                    source=cnode.id,
                    target=dnode.id,
                    abs_uri=str(RDFS.label),
                    rel_uri=kgns.get_rel_uri(RDFS.label),
                )
            )

        for source, target, predicate in cpa:
            unode = nodes[source]
            vnode = nodes[target]

            if source not in nodemap:
                nodemap[source] = sm.add_node(unode)
            if target not in nodemap:
                nodemap[target] = sm.add_node(vnode)

        # detect and handle untype source column node
        remove_rel_from_nodes = set()

        for source, target, predicate in cpa:
            unode = nodes[source]
            if isinstance(unode, DataNode):
                if unode.col_index not in classmap:
                    # discover untyped source column node
                    if on_untype_source_column_node == "create-class":
                        # this data node has an outgoing edge, but it's untyped
                        # so we create an entity class to represent its type
                        cnode_id = sm.add_node(
                            ClassNode(
                                abs_uri=self.kgns.entity_uri,
                                rel_uri=kgns.get_rel_uri(self.kgns.entity_uri),
                                readable_label=self.kgns.entity_label,
                            )
                        )
                        classmap[unode.col_index] = cnode_id
                        sm.add_edge(
                            Edge(
                                source=cnode_id,
                                target=unode.id,
                                abs_uri=str(RDFS.label),
                                rel_uri=kgns.get_rel_uri(RDFS.label),
                            )
                        )
                    else:
                        assert on_untype_source_column_node == "remove-link"
                        vnode = nodes[target]

                        if (
                            isinstance(vnode, ClassNode)
                            and vnode.abs_uri == kgns.statement_uri
                        ):
                            # this is a statement node, so we need to remove link
                            # from the statement node too because statement node
                            # can't live without the source
                            remove_rel_from_nodes.add(target)
                        remove_rel_from_nodes.add(source)

        if len(remove_rel_from_nodes) > 0:
            assert on_untype_source_column_node == "remove-link"
            cpa = [
                (source, target, predicate)
                for source, target, predicate in cpa
                if source not in remove_rel_from_nodes
            ]

        for source, target, predicate in cpa:
            unode = nodes[source]
            vnode = nodes[target]

            # if source not in nodemap:
            #     nodemap[source] = sm.add_node(unode)
            # if target not in nodemap:
            #     nodemap[target] = sm.add_node(vnode)

            if isinstance(unode, DataNode):
                # outgoing edge is from a class node instead of a data node
                assert unode.col_index in classmap
                # if unode.col_index not in classmap:
                #     # this data node has an outgoing edge, but it's untyped
                #     # so we create an entity class to represent its type
                #     curl = kgns.get_entity_abs_uri(self.ENTITY_ID)
                #     cnode_id = sm.add_node(
                #         ClassNode(
                #             abs_uri=curl,
                #             rel_uri=kgns.get_entity_rel_uri(self.ENTITY_ID),
                #             readable_label=self.kgns.entity_label,
                #         )
                #     )
                #     classmap[unode.col_index] = cnode_id
                #     sm.add_edge(
                #         Edge(
                #             source=cnode_id,
                #             target=unode.id,
                #             abs_uri=str(RDFS.label),
                #             rel_uri=kgns.get_rel_uri(RDFS.label),
                #         )
                #     )
                suid = classmap[unode.col_index]
                source = sm.get_node(suid)
            else:
                source = unode

            if isinstance(vnode, DataNode):
                # if this is an entity column, the link should map to its class
                if vnode.col_index in classmap:
                    target = sm.get_node(classmap[vnode.col_index])
                else:
                    target = vnode
            else:
                target = vnode

            sm.add_edge(
                Edge(
                    source=source.id,
                    target=target.id,
                    abs_uri=(tmp_abs_uri := kgns.id_to_uri(predicate)),
                    rel_uri=kgns.get_rel_uri(tmp_abs_uri),
                    readable_label=self.get_pnode_label(predicate),
                )
            )

        return sm

    def create_sm_from_column_rels(
        self,
        cpa: Sequence[tuple[int, int, str]],
        cta: Mapping[int, str],
        columns: list[str] | dict[int, str],
        validate: bool = True,
    ):
        nodes = {}
        if isinstance(columns, list):
            for ci, cname in enumerate(columns):
                nodes[ci] = DataNode(ci, cname)
        else:
            for ci, cname in columns.items():
                nodes[ci] = DataNode(ci, cname)
        return self.create_sm(nodes, cpa, cta, validate)

    def convert_sm_to_column_rels(
        self, sm: SemanticModel
    ) -> tuple[dict[int, str], list[tuple[int, int, str]]]:
        raise NotImplementedError()



import os


class AppConfig:
    instance = None

    @staticmethod
    def get_instance():
        if AppConfig.instance is None:
            AppConfig.instance = AppConfig()
        return AppConfig.instance

    @property
    def is_canrank_verbose(self) -> bool:
        return os.environ.get("CANRANK_VERBOSE") == "1"

    @property
    def is_cache_enable(self) -> bool:
        return os.environ.get("ENABLE_CACHE") == "1"

    @property
    def is_ray_disable(self) -> bool:
        return os.environ.get("DISABLE_RAY") == "1"



import gramsplus.core.algorithms as algorithms
import gramsplus.core.literal_matchers as literal_matchers
import gramsplus.core.models as models

__all__ = ["algorithms", "literal_matchers", "models"]



from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Callable, Literal, Mapping, Union

import orjson
from hugedict.misc import identity
from hugedict.sqlitedict import SqliteDict, SqliteDictFieldType
from kgdata.models.ont_class import OntologyClass
from kgdata.models.ont_property import OntologyProperty
from loguru import logger
from sm.evaluation.hierarchy_scoring_fn import (
    INF_DISTANCE,
    MAX_ANCESTOR_DISTANCE,
    MAX_DESCENDANT_DISTANCE,
    HierarchyScoringFn,
)
from sm.namespaces.namespace import KnowledgeGraphNamespace
from sm.namespaces.wikidata import WikidataNamespace

MAX_DISTANCE = max(MAX_ANCESTOR_DISTANCE, MAX_DESCENDANT_DISTANCE)


class ItemType(str, Enum):
    CLASS = "class"
    PROPERTY = "property"


class SqliteItemDistance(SqliteDict[str, int]):
    def __init__(
        self,
        path: Union[str, Path],
        items: Mapping[str, OntologyClass] | Mapping[str, OntologyProperty],
        kgns: KnowledgeGraphNamespace,
        timeout: float = 5.0,
    ):
        super().__init__(
            path,
            keytype=SqliteDictFieldType.str,
            ser_value=identity,
            deser_value=identity,
            valuetype=SqliteDictFieldType.int,
            timeout=timeout,
        )

        logger.warning("SqliteItemDistance is deprecated, use KGItemDistance instead")

        self._items = items
        self._cache_distance = {}
        self.uri2id = kgns.uri_to_id
        self.is_valid_uri = kgns.is_uri_in_ns
        self.is_valid_id = kgns.is_id

    def get_distance(self, pred_item: str, target_item: str) -> int:
        """Calculate distance between two items. Positive if pred_id is the ancestor of target_id, negative otherwise."""
        if pred_item == target_item:
            return 0

        if (pred_item, target_item) not in self._cache_distance:
            if self.is_valid_uri(pred_item):
                pred_id = self.uri2id(pred_item)
            else:
                assert self.is_valid_id(pred_item), pred_item
                pred_id = pred_item

            if self.is_valid_uri(target_item):
                target_id = self.uri2id(target_item)
            else:
                assert self.is_valid_id(target_item), target_item
                target_id = target_item

            key = orjson.dumps([pred_id, target_id]).decode()
            if key not in self:
                distance = self._calculate_distance(pred_id, target_id)
                self[key] = distance
            self._cache_distance[pred_item, target_item] = self[key]
        return self._cache_distance[pred_item, target_item]

    def batch_get_distance(self, items: list[tuple[str, str]]) -> list[int]:
        distances = []
        newdistances = []
        for pred_item, target_item in items:
            if pred_item == target_item:
                distances.append(0)
                continue

            if (pred_item, target_item) not in self._cache_distance:
                if self.is_valid_uri(pred_item):
                    pred_id = self.uri2id(pred_item)
                else:
                    assert self.is_valid_id(pred_item), pred_item
                    pred_id = pred_item

                if self.is_valid_uri(target_item):
                    target_id = self.uri2id(target_item)
                else:
                    assert self.is_valid_id(target_item), target_item
                    target_id = target_item

                key = orjson.dumps([pred_id, target_id]).decode()
                if key not in self:
                    distance = self._calculate_distance(pred_id, target_id)
                    newdistances.append((key, distance))
                else:
                    distance = self[key]
                self._cache_distance[pred_item, target_item] = distance

            distances.append(self._cache_distance[pred_item, target_item])
        if len(newdistances) > 0:
            self.batch_insert(newdistances)
        return distances

    def _calculate_distance(self, pred_id: str, target_id: str) -> int:
        """Calculate distance between two items. Positive if pred_id is the ancestor of target_id, negative otherwise."""
        targ_obj = self._items[target_id]

        if pred_id in targ_obj.ancestors:
            # predicted item is the ancestor of the target
            return self._calculate_distance_to_ancestors(targ_obj, pred_id)

        pred_obj = self._items[pred_id]
        if target_id in pred_obj.ancestors:
            # predicted item is the descendant of the target
            return -self._calculate_distance_to_ancestors(pred_obj, target_id)

        return INF_DISTANCE

    def _calculate_distance_to_ancestors(
        self, source: OntologyClass | OntologyProperty, target_id: str
    ) -> int:
        visited = {}
        stack: list[tuple[str, int]] = [(uid, 1) for uid in source.parents]
        while len(stack) > 0:
            uid, distance = stack.pop()
            if uid in visited and distance >= visited[uid]:
                # we have visited this node before and since last time we visit
                # the previous route is shorter, so we don't need to visit it again
                continue

            visited[uid] = distance
            u = self._items[uid]
            if target_id == uid or target_id not in u.ancestors:
                # this is a dead-end path, don't need to explore
                continue
            for parent_id in u.parents:
                if distance < MAX_DISTANCE:
                    # still within the distance limit, we can explore further
                    stack.append((parent_id, distance + 1))
        # we may have not visited the target node because its too far away (> MAX_DISTANCE).
        return visited.get(target_id, INF_DISTANCE)


class KGItemDistance:
    """Calculate distance between items from KGData"""

    def __init__(
        self,
        items: Mapping[str, OntologyClass] | Mapping[str, OntologyProperty],
        kgns: KnowledgeGraphNamespace,
    ):
        self.items = items
        self.uri2id = kgns.uri_to_id
        self.is_valid_id = kgns.is_id
        self.is_valid_uri = kgns.is_uri_in_ns

    def get_distance(self, pred_item: str, target_item: str) -> int:
        if pred_item == target_item:
            return 0

        if self.is_valid_uri(pred_item):
            pred_id = self.uri2id(pred_item)
        else:
            assert self.is_valid_id(pred_item), pred_item
            pred_id = pred_item

        if self.is_valid_uri(target_item):
            target_id = self.uri2id(target_item)
        else:
            assert self.is_valid_id(target_item), target_item
            target_id = target_item

        pred_obj = self.items[pred_id]
        target_obj = self.items[target_id]

        if pred_id in target_obj.ancestors:
            return target_obj.ancestors[pred_id]
        if target_id in pred_obj.ancestors:
            return -pred_obj.ancestors[target_id]
        return INF_DISTANCE

    def batch_get_distance(self, items: list[tuple[str, str]]) -> list[int]:
        return [
            self.get_distance(pred_item, target_item)
            for pred_item, target_item in items
        ]


def get_hierarchy_scoring_fn(
    items: Mapping[str, OntologyClass] | Mapping[str, OntologyProperty],
    kgns: KnowledgeGraphNamespace,
) -> HierarchyScoringFn:
    return HierarchyScoringFn(KGItemDistance(items, kgns))



from __future__ import annotations

from collections import defaultdict
from copy import copy
from operator import itemgetter
from typing import Literal, Optional, cast

from gramsplus.semanticmodeling.postprocessing.cgraph import (
    CGEdge,
    CGEdgeTriple,
    CGNode,
    CGraph,
)
from gramsplus.semanticmodeling.postprocessing.common import (
    add_context,
    ensure_valid_statements,
)
from gramsplus.semanticmodeling.postprocessing.config import PostprocessingConfig
from gramsplus.semanticmodeling.postprocessing.interface import (
    EdgeProb,
    NodeProb,
    PostprocessingFn,
    PostprocessingResult,
)
from graph.retworkx.api import dag_longest_path, has_cycle
from sm.dataset import FullTable
from steiner_tree.bank.solver import PSEUDO_ROOT_ID, BankSolver
from steiner_tree.bank.struct import (
    BankEdge,
    BankGraph,
    BankNode,
    NoSingleRootException,
    Solution,
    UpwardTraversal,
)


class SteinerTreeFn(PostprocessingFn):
    def __init__(
        self,
        table: FullTable,
        cg: CGraph,
        cpa_threshold: float,
        cta_threshold: float,
        cta_score_offset: float,
    ):
        self.table = table
        self.cpa_threshold = cpa_threshold
        self.cta_threshold = cta_threshold + cta_score_offset
        self.cg = cg

    def __call__(
        self, node_probs: NodeProb, edge_probs: EdgeProb
    ) -> PostprocessingResult:
        st = SteinerTree(
            self.table,
            self.cg,
            edge_probs,
            self.cpa_threshold,
            additional_terminal_nodes=None,
        )
        pred_cg = st.get_result()

        edges = [(e.source, e.target, e.key) for e in pred_cg.iter_edges()]
        nodes = {}
        for colindex, utypes in node_probs.items():
            ctype, cscore = max(utypes.items(), key=itemgetter(1))
            if cscore >= self.cta_threshold:
                nodes[colindex] = ctype
        return PostprocessingResult(edges=edges, nodes=nodes)


class SteinerTree:
    def __init__(
        self,
        table: FullTable,
        cg: CGraph,
        edge_probs: dict[CGEdgeTriple, float],
        threshold: float,
        additional_terminal_nodes: Optional[list[str]] = None,
    ):
        self.table = table
        self.cg = cg
        self.edge_probs = edge_probs
        self.threshold = threshold

        self.top_k_st = 50
        self.top_k_path = 50

        # extra terminal nodes that the tree should have, usually used in
        # interactive modeling where users add some entity nodes in their model
        self.additional_terminal_nodes = additional_terminal_nodes

    def get_result(self) -> CGraph:
        """Select edges that forms a tree"""
        edge_probs = {e: p for e, p in self.edge_probs.items() if p >= self.threshold}

        # first step is to remove dangling statements and standalone nodes
        subcg = self.cg.subgraph_from_edge_triples(edge_probs.keys())
        subcg.remove_dangling_statement()
        subcg.remove_standalone_nodes()

        # terminal nodes are columns node, entity nodes are adding later
        terminal_nodes = {u.id for u in subcg.iter_nodes() if u.is_column_node}
        if self.additional_terminal_nodes is not None:
            terminal_nodes.update(self.additional_terminal_nodes)
        if len(terminal_nodes) == 0:
            # if the tree does not have any column, we essentially don't predict anything
            # so we return an empty graph
            return CGraph()

        # add pseudo root
        edge_weights = {e: 1.0 / p for e, p in edge_probs.items()}  # p > 0.5

        solver = BankSteinerTree(
            original_graph=subcg,
            terminal_nodes=terminal_nodes,
            top_k_st=self.top_k_st,
            top_k_path=self.top_k_path,
            weight_fn=lambda e: edge_weights[e.source, e.target, e.key],
            solution_cmp_fn=self.compare_solutions,
            invalid_roots={u.id for u in subcg.nodes() if u.is_statement_node},
            allow_shorten_graph=False,
        )
        trees, _solutions = solver.run()
        if len(trees) == 0:
            return CGraph()

        tree = cast(CGraph, trees[0])
        tree.remove_dangling_statement()

        if PostprocessingConfig.INCLUDE_CONTEXT:
            add_context(subcg, tree, edge_probs)

        # add back statement property if missing into to ensure a correct model
        ensure_valid_statements(subcg, tree, create_if_not_exists=False)

        # fmt: off
        # from graph.viz.graphviz import draw
        # draw(graph=tree, filename="/tmp/graphviz/st204.png", **CGGraph.graphviz_props())
        # draw(graph=tree, filename="/tmp/graphviz/g25.png", **CGGraph.graphviz_props())
        # fmt: on

        return tree

    def compare_solutions(self, a: Solution, b: Solution) -> int:
        """Comparing two solutions, -1 (smaller) means a better solution -- we are solving minimum steiner tree"""
        a_weight = a.weight / max(a.num_edges, 1)
        b_weight = b.weight / max(b.num_edges, 1)

        if a_weight < b_weight:
            return -1
        if a_weight > b_weight:
            return 1
        # equal weight, prefer the one with shorter depth
        # TODO: there is a case where a statement has a qualifier
        # it makes the tree without source node shorter than the one with source node
        # however, the one with the source node is better...
        # so the steiner tree should have an option to say like if we pick this path, we must
        # select the source node to avoid these issue...

        # for a dirty fix, perhaps we can say, the one with a dangling statement is infinitly bad
        if not hasattr(a, "has_no_source_statement"):
            has_no_source_statement = False
            for u in a.graph.iter_nodes():
                if u.id == PSEUDO_ROOT_ID:
                    continue
                cgu = self.cg.get_node(u.id)
                if cgu.is_statement_node and a.graph.in_degree(u.id) == 0:
                    has_no_source_statement = True
                    break
            setattr(a, "has_no_source_statement", has_no_source_statement)

        if not hasattr(b, "has_no_source_statement"):
            has_no_source_statement = False
            for u in b.graph.iter_nodes():
                if u.id == PSEUDO_ROOT_ID:
                    continue
                cgu = self.cg.get_node(u.id)
                if cgu.is_statement_node and b.graph.in_degree(u.id) == 0:
                    has_no_source_statement = True
                    break
            setattr(b, "has_no_source_statement", has_no_source_statement)

        a_has_no_source = getattr(a, "has_no_source_statement")
        b_has_no_source = getattr(b, "has_no_source_statement")
        if not a_has_no_source and b_has_no_source:
            return -1
        if a_has_no_source and not b_has_no_source:
            return 1

        if not hasattr(a, "depth"):
            setattr(a, "depth", len(dag_longest_path(a.graph)))
        if not hasattr(b, "depth"):
            setattr(b, "depth", len(dag_longest_path(b.graph)))
        return getattr(a, "depth") - getattr(b, "depth")


class BankSteinerTree(BankSolver[CGNode, CGEdge]):
    ADD_MISSING_STATEMENT_PROPS = True

    def _solve(
        self, g: BankGraph, terminal_nodes: set[str], top_k_st: int, top_k_path: int
    ):
        """Override the main algorithm to handle statement"""
        roots = {u.id for u in g.iter_nodes() if u.id not in self.invalid_roots}

        attr_visit_hists: list[tuple[str, UpwardTraversal]] = []
        # to ensure the order
        for uid in list(sorted(terminal_nodes)):
            visit_hist = UpwardTraversal.top_k_beamsearch(g, uid, top_k_path)
            roots = roots.intersection(visit_hist.paths.keys())
            attr_visit_hists.append((uid, visit_hist))

        if len(roots) == 0:
            raise NoSingleRootException()

        # to ensure the order again & remove randomness
        roots = sorted(roots)

        # merge the paths using beam search
        results = []
        for root in roots:
            current_states = []
            uid, visit_hist = attr_visit_hists[0]
            for path in visit_hist.paths[root]:
                pg = BankGraph()
                if len(path.path) > 0:
                    assert uid == path.path[0].target
                pg.add_node(BankNode(uid))
                for e in path.path:
                    pg.add_node(BankNode(e.source))
                    pg.add_edge(e.clone())

                self.add_missing_statement(pg)
                current_states.append(pg)

            if len(current_states) > top_k_st:
                current_states = [
                    _s.graph for _s in self._sort_solutions(current_states)[:top_k_st]
                ]

            for uid, visit_hist in attr_visit_hists[1:]:
                next_states = []
                for state in current_states:
                    for path in visit_hist.paths[root]:
                        pg = state.copy()
                        if len(path.path) > 0:
                            assert uid == path.path[0].target
                        if not pg.has_node(uid):
                            pg.add_node(BankNode(uid))
                        for e in path.path:
                            if not pg.has_node(e.source):
                                pg.add_node(BankNode(id=e.source))
                            # TODO: here we don't check by edge_key because we may create another edge of different key
                            # hope this new path has been exploited before.
                            if not pg.has_edges_between_nodes(e.source, e.target):
                                pg.add_edge(e.clone())
                        self.add_missing_statement(pg)

                        # if there are more than path between two nodes within
                        # two hop, we'll select one
                        update_graph = False
                        for n in pg.iter_nodes():
                            if pg.in_degree(n.id) >= 2:
                                grand_parents: dict[
                                    str, list[tuple[BankEdge, ...]]
                                ] = defaultdict(list)
                                for inedge in pg.in_edges(n.id):
                                    grand_parents[inedge.source].append((inedge,))
                                    for grand_inedge in pg.in_edges(inedge.source):
                                        grand_parents[grand_inedge.source].append(
                                            (grand_inedge, inedge)
                                        )

                                for grand_parent, edges in grand_parents.items():
                                    if len(edges) > 1:
                                        # we need to select one path from this grand parent to the rest
                                        # they have the same length, so we select the one has smaller weight
                                        edges = sorted(
                                            edges,
                                            key=lambda x: x[0].weight + x[1].weight
                                            if len(x) == 2
                                            else x[0].weight * 2,
                                        )

                                        for lst in edges[1:]:
                                            for edge in lst:
                                                # TODO: handle removing edges multiple times
                                                try:
                                                    pg.remove_edge(edge.id)
                                                except:
                                                    continue
                                        update_graph = True
                        if update_graph:
                            for n in pg.nodes():
                                if pg.in_degree(n.id) == 0 and pg.out_degree(n.id) == 0:
                                    pg.remove_node(n.id)
                        # after add a path to the graph, it can create new cycle, detect and fix it
                        if has_cycle(pg):
                            # we can show that if the graph contain cycle, there is a better path
                            # so no need to try to break cycles as below
                            # cycles_iter = [(uid, vid) for uid, vid, eid, orien in cycles_iter]
                            # for _g in self._break_cycles(root, pg, cycles_iter):
                            #     next_states.append(_g)
                            pass
                        else:
                            next_states.append(pg)

                        # the output graph should not have parallel edges
                        assert not pg.has_parallel_edges()
                if len(next_states) > top_k_st:
                    next_states = [
                        _s.graph for _s in self._sort_solutions(next_states)[:top_k_st]
                    ]
                # assert all(_.check_integrity() for _ in next_states)
                current_states = next_states
                # cgs = [g for g in next_states if len(list(nx.simple_cycles(g))) > 0]
                # nx.draw_networkx(cgs[0]); plt.show()
                # nx.draw(cgs[0]); plt.show()
            results += current_states

        return self._sort_solutions(results)

    def add_missing_statement(self, g: BankGraph):
        if not self.ADD_MISSING_STATEMENT_PROPS:
            return

        cg = cast(CGraph, self.original_graph)
        for uprime in g.iter_nodes():
            if uprime.id == PSEUDO_ROOT_ID:
                continue
            u = cg.get_node(uprime.id)
            if u.is_statement_node:
                (inedge,) = cg.in_edges(u.id)
                if all(outedge.key != inedge.key for outedge in g.out_edges(u.id)):
                    # no statement property, add it back
                    # sometimes the baseline don't have main statement
                    tmp = [
                        outedge
                        for outedge in self.graph.out_edges(u.id)
                        if outedge.key == inedge.key
                    ]
                    if len(tmp) == 0:
                        continue
                    (main_prop,) = tmp
                    if not g.has_node(main_prop.target):
                        target_id = g.add_node(
                            copy(self.graph.get_node(main_prop.target))
                        )
                        assert target_id == main_prop.target
                    g.add_edge(main_prop.clone())



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Optional, Sequence, TypeVar, Union

from sm.evaluation.prelude import PrecisionRecallF1Support
from sm.evaluation.utils import PrecisionRecallF1
from sm.misc.matrix import Matrix


@dataclass
class ConfusionMatrix:
    tp: int = 0  # true positive
    fn: int = 0  # false negative
    fp: int = 0  # false positive
    tn: int = 0  # true negative

    def total(self):
        return self.tp + self.fn + self.fp + self.tn

    def precision(self):
        if self.total() == 0:
            return 1.0
        # the system must always predict something. NO_ENTITY is treated as a prediction
        if self.tp + self.fp > 0:
            return self.tp / (self.tp + self.fp)
        return 0.0

    def recall(self):
        if self.total() == 0:
            return 1.0
        # the system must always predict something. NO_ENTITY is treated as a prediction
        if self.tp + self.fn > 0:
            return self.tp / (self.tp + self.fn)
        return 0.0

    def f1(self):
        p = self.precision()
        r = self.recall()
        if p + r == 0:
            return 0.0
        return 2 * p * r / (p + r)

    def get_precision_recall_f1(self) -> PrecisionRecallF1:
        return PrecisionRecallF1(self.precision(), self.recall(), self.f1())

    def __add__(self, another: ConfusionMatrix):
        return ConfusionMatrix(
            self.tp + another.tp,
            self.fn + another.fn,
            self.fp + another.fp,
            self.tn + another.tn,
        )


def inkb_eval_table(
    gold_ents: Matrix[set[str]],
    pred_ents: Matrix[list[str]],
    k: Optional[Union[int, Sequence[Optional[int]]]] = None,
) -> tuple[dict[str, PrecisionRecallF1Support], dict[str, ConfusionMatrix]]:
    """Evaluate entity linking performance on a table using InKB metrics. As defined in Gerbil framework, InKB metrics
    only consider entities in KB as correct entities, no entities (incorrect mentions) or entities outside of KB (NIL) are
    consider incorrect and eliminated. In other words, only mentions with valid KB entities are used for evaluation.

    NOTE: This evaluation assumes that no pred entity is still a prediction (NO_ENTITY). Therefore, the system will never be able to
    achieve 1.0 precision by not predicting anything.

    Args:
        gold_ents: Matrix of gold entities for each cell. If the set of entities of a cell is empty, the
            cell does not contain any entity (no entity in KB).
        pred_ents: Matrix of predicted entities for each cell. If the list of candidate entities of a
            cell is empty, we do not predict any entity for the cell. Candidates in the list are sorted by
            their likelihood in descending order.
    """
    nrows, ncols = gold_ents.shape()
    name2k: dict[str, Optional[int]] = {"perf@all": None}
    if k is not None:
        for ki in k if isinstance(k, Sequence) else [k]:
            if ki is not None:
                name2k[f"perf@{ki}"] = ki

    confusion_matrices = {name: ConfusionMatrix() for name in name2k}

    for i in range(nrows):
        for j in range(ncols):
            if len(gold_ents[i, j]) == 0:
                # no entity in gold, but we predict something, this is obviously wrong
                # but since we ignore no entity or incorrect mentions, we do not count
                # this as a false positive
                continue

                # if len(pred_ents[i, j]) == 0:
                #     # no entity in gold, no entity in pred, so no example
                #     continue
                # for name in name2k:
                #     confusion_matrices[name].fp += 1
            else:
                ytrue = gold_ents[i, j]
                ypreds = pred_ents[i, j]

                for name, k in name2k.items():
                    if len(ytrue.intersection(ypreds[:k])) > 0:
                        confusion_matrices[name].tp += 1
                    else:
                        confusion_matrices[name].fn += 1
                        # so recall = precision
                        confusion_matrices[name].fp += 1

    output = {}
    for name, ki in name2k.items():
        cm = confusion_matrices[name]
        output[name] = PrecisionRecallF1Support(
            recall=cm.recall(),
            precision=cm.precision(),
            f1=cm.f1(),
            support=cm.tp + cm.fn,
        )

    return output, confusion_matrices


T = TypeVar("T", str, int, tuple[str, ...])


def inkb_mrr(
    ytrue: Sequence[T] | Sequence[set[T]] | Matrix[set[T]],
    ypreds: Sequence[Sequence[T]] | Matrix[Sequence[T]],
    is_in_groundtruth: Optional[Callable[[T, set[T]], bool]] = None,
) -> float:
    """Calculate MRR for a list of queries. Optionally support passing a matrix.

    This follow the GERBIL InKB metrics. Only mentions with valid KB entities are used for evaluation.
    """
    if is_in_groundtruth is None:
        is_in_groundtruth = lambda ypred, ytrues: ypred in ytrues

    if isinstance(ytrue, Matrix):
        assert isinstance(ypreds, Matrix)
        nrows, ncols = ytrue.shape()
        out = 0.0
        n = 0
        for ri in range(nrows):
            for ci in range(ncols):
                y = ytrue[ri, ci]
                if len(y) == 0:
                    # skip mentions with no entity
                    continue

                out += next(
                    (
                        1 / j
                        for j, ypred in enumerate(ypreds[ri, ci], start=1)
                        if is_in_groundtruth(ypred, y)
                    ),
                    0,
                )
                n += 1

        return out / max(n, 1)

    if len(ytrue) == 0:
        return 0.0

    out = 0.0

    assert isinstance(ypreds, Sequence)

    if isinstance(ytrue[0], str):
        for i in range(len(ytrue)):
            y = ytrue[i]
            out += next(
                (
                    1 / j
                    for j, ypred in enumerate(ypreds[i], start=1)
                    if is_in_groundtruth(ypred, {y})  # type: ignore
                ),
                0,
            )
    else:
        for i in range(len(ytrue)):
            y = ytrue[i]
            assert len(y) > 0  # type: ignore
            out += next(
                (
                    1 / j
                    for j, ypred in enumerate(ypreds[i], start=1)
                    if is_in_groundtruth(ypred, y)  # type: ignore
                ),
                0,
            )
    return out / max(len(ytrue), 1)



from __future__ import annotations

from copy import copy
from operator import itemgetter
from typing import TypedDict

from gp.semanticmodeling.postprocessing.cgraph import CGEdge, CGraph

_AddContextPathType = TypedDict(
    "_AddContextPathType", {"path": tuple[CGEdge, CGEdge], "score": float}
)


def ensure_valid_statements(
    original_cg: CGraph,
    pruned_cg: CGraph,
    create_if_not_exists: bool = False,
    safe_creation: bool = True,
):
    """Ensure that the pruned graph has valid statements, which have statement property.

    NOTE: this function may modify the pruned graph.

    Args:
        original_cg: the original candidate graph
        pruned_cg: the pruned candidate graph -- must have the same id as the original cg
        create_if_not_exists: if True, create a new target of statement property if it does not exist
        safe_creation: if True, we check to make sure the new statement target is not in the graph
    """
    for s in pruned_cg.nodes():
        if not s.is_statement_node:
            continue

        (inedge,) = pruned_cg.in_edges(s.id)
        for outedge in original_cg.out_edges(s.id):
            if outedge.key == inedge.key:
                # there is a statement prop in the original candidate graph
                # if it is not in the tree, we need to add it back
                if pruned_cg.has_edge_between_nodes(
                    outedge.source, outedge.target, outedge.key
                ):
                    continue

                if not create_if_not_exists:
                    raise Exception(
                        f"The property of the statement {s.id} is not in the graph"
                    )

                if safe_creation:
                    assert not pruned_cg.has_node(
                        outedge.target
                    ), "The target should not in the grpah as we do not have the link"
                    target = copy(original_cg.get_node(outedge.target))
                    assert (
                        target.is_entity_node or target.is_literal_node
                    ), "The only reason why we don't have statement value is it is an entity/literal"
                    pruned_cg.add_node(target)
                elif not pruned_cg.has_node(outedge.target):
                    target = copy(original_cg.get_node(outedge.target))
                    pruned_cg.add_node(target)

                pruned_cg.add_edge(copy(outedge))


def add_context(
    origin_cg: CGraph,
    pruned_cg: CGraph,
    edge_probs: dict[tuple[str, str, str], float],
):
    """Add back the context node or entity that are appeared in the original graph but not in the pruned graph.

    Note: this function may modify the pruned graph.

    Args:
        origin_cg: the original candidate graph
        pruned_cg: the pruned candidate graph -- must have the same id as the original cg
        edge_probs: containing edges' probabilities
    """
    for v in origin_cg.iter_nodes():
        if (
            not (v.is_entity_node or v.is_literal_node)
            or not v.is_in_context
            or pruned_cg.has_node(v.id)
        ):
            continue

        # find the paths that connect the vid to the tree and select the one with highest score and do not create cycle
        paths: list[_AddContextPathType] = []
        for sv_edge in origin_cg.in_edges(v.id):
            if sv_edge.key == "P31":  # ignore instanceof
                continue
            for us_edge in origin_cg.in_edges(sv_edge.source):
                if not pruned_cg.has_node(us_edge.source):
                    continue
                paths.append(
                    {
                        "path": (
                            us_edge,
                            sv_edge,
                        ),
                        "score": edge_probs[
                            (us_edge.source, us_edge.target, us_edge.key)
                        ]
                        + edge_probs[(sv_edge.source, sv_edge.target, sv_edge.key)],
                    }
                )

        paths = sorted(paths, key=itemgetter("score"), reverse=True)

        if len(paths) == 0:
            continue

        us_edge, sv_edge = paths[0]["path"]
        pruned_cg.add_node(copy(v))
        if not pruned_cg.has_node(sv_edge.source):
            s = origin_cg.get_node(sv_edge.source)
            pruned_cg.add_node(copy(s))
        assert not pruned_cg.has_edge_between_nodes(
            sv_edge.source, sv_edge.target, sv_edge.key
        )
        pruned_cg.add_edge(copy(sv_edge))

        if not pruned_cg.has_edge_between_nodes(
            us_edge.source, us_edge.target, us_edge.key
        ):
            pruned_cg.add_edge(copy(us_edge))



from typing import Dict, List, Optional, Tuple

import networkx as nx
from graph.interface import BaseNode
from sm.dataset import FullTable

from gramsplus.semanticmodeling.postprocessing.cgraph import (
    CGEdge,
    CGEdgeTriple,
    CGNode,
    CGraph,
)
from gramsplus.semanticmodeling.postprocessing.common import (
    add_context,
    ensure_valid_statements,
)
from gramsplus.semanticmodeling.postprocessing.config import PostprocessingConfig


class MinimumArborescence:
    def __init__(
        self,
        table: FullTable,
        cg: CGraph,
        edge_probs: Dict[CGEdgeTriple, float],
        threshold: float,
    ):
        self.table = table
        self.cg = cg
        self.edge_probs = edge_probs
        self.threshold = threshold

    def add_pseudo_root(self, cg: CGraph, pseudo_root_id: str = "__root__") -> CGraph:
        """Add a pseudo root node and edges from non-statement nodes to the pseudo root
        so that we can always have a directed rooted tree

        Args:
            cg: the graph to impute
            pseudo_root_id: the id of the pseudo root node
        """
        newcg = cg.copy()
        assert newcg.has_node(pseudo_root_id) is False

        newcg.add_node(
            CGNode(
                id=pseudo_root_id,
                is_statement_node=False,
                is_column_node=False,
                is_entity_node=False,
                is_literal_node=False,
                is_in_context=False,
                column_index=False,
            )
        )

        for node in cg.iter_nodes():
            if node.is_statement_node:
                continue

            newcg.add_edge(
                CGEdge(id=-1, source=pseudo_root_id, target=node.id, key="null")
            )

        return newcg

    def get_result(self) -> CGraph:
        """Select edges that forms a tree"""
        edge_probs = {e: p for e, p in self.edge_probs.items() if p >= self.threshold}
        subcg = self.cg.subgraph_from_edge_triples(edge_probs.keys())
        subcg.remove_dangling_statement()
        subcg.remove_standalone_nodes()

        if subcg.num_edges() == 0:
            return subcg

        # for each node that is not a statement, add a common root node and edges to it
        pseudo_root_id = "__root__"
        new_subcg = self.add_pseudo_root(subcg, pseudo_root_id)

        # find a minimum arborescence
        nxcg = nx.MultiDiGraph()
        edge_weights = {e: 1.0 / p for e, p in edge_probs.items()}  # p > 0.5
        total_weight = sum(edge_weights.values())

        for edge in new_subcg.out_edges(pseudo_root_id):
            edge_weights[edge.source, edge.target, edge.key] = total_weight + 1

        for edge in new_subcg.iter_edges():
            nxcg.add_edge(
                edge.source,
                edge.target,
                key=edge.key,
                weight=edge_weights[edge.source, edge.target, edge.key],
                edgekey=edge.key,
            )

        try:
            nxtree = nx.algorithms.tree.branchings.minimum_spanning_arborescence(
                nxcg, attr="weight", preserve_attrs=True
            )
        except nx.NetworkXException as e:
            assert str(e) == "No minimum spanning arborescence in G.", str(e)
            raise NoMinimumArborescenceException() from e

        # remove the pseudo root node
        nxtree.remove_node(pseudo_root_id)

        # convert tree to nxtree
        tree_edges = [
            (uid, vid, ekey) for uid, vid, ekey in nxtree.edges(data="edgekey")
        ]
        tree = subcg.subgraph_from_edge_triples(tree_edges)
        tree.remove_dangling_statement()

        clean_entities(tree)

        if PostprocessingConfig.INCLUDE_CONTEXT:
            add_context(subcg, tree, edge_probs)

        ensure_valid_statements(subcg, tree, create_if_not_exists=True)

        # fmt: off
        # from graph.viz.graphviz import draw
        # draw(graph=tree, filename="/tmp/graphviz/a204.png", **CGraph.graphviz_props())
        # draw(graph=tree, filename="/tmp/graphviz/g25.png", **CGraph.graphviz_props())
        # fmt: on
        return tree


class NoMinimumArborescenceException(Exception):
    pass


def clean_entities(tree: CGraph):
    for u in tree.nodes():
        if not (u.is_entity_node or u.is_literal_node):
            continue

        if u.is_in_context:
            continue

        if tree.in_degree(u.id) == 0:
            if tree.out_degree(u.id) == 1:
                tree.remove_node(u.id)
            continue

        (svedge,) = tree.in_edges(u.id)
        stmt = tree.get_node(svedge.source)
        assert stmt.is_statement_node

        (usedge,) = tree.in_edges(stmt.id)
        if usedge.key != svedge.key:
            # qualifier
            tree.remove_node(u.id)
            continue

        # property
        if tree.out_degree(stmt.id) == 1:
            tree.remove_node(u.id)
            tree.remove_node(stmt.id)

    tree.remove_dangling_statement()



from __future__ import annotations

import re
from dataclasses import dataclass, field
from functools import lru_cache
from operator import itemgetter
from pathlib import Path
from typing import Mapping, Optional, Sequence

import serde.csv
from gp.actors.data._db import KGDB, DBActor

# from gp.actors.data.data_autolabel import AutoLabeledTable
from gp.misc.evaluation.sm_wikidata import SemanticModelHelper
from libactor.actor import Actor
from libactor.cache import BackendFactory, IdentObj, cache
from rdflib import RDFS
from sm.dataset import Dataset, Example, FullTable
from sm.namespaces.namespace import KnowledgeGraphNamespace
from sm.namespaces.utils import KGName, get_kgns
from sm.prelude import M, O
from smml.dataset_helper import DatasetList, DatasetQuery

try:
    from sm_datasets.datasets import Datasets
except ImportError:
    Datasets = None


@dataclass
class DataActorArgs:
    skip_unk_ont_ent: bool = field(
        default=False,
        metadata={
            "help": "Skip examples with unknown ontology entities (e.g., Q1234567)"
        },
    )
    skip_no_sm: bool = field(
        default=False,
        metadata={"help": "Skip examples without semantic models"},
    )
    include_dataset_name: bool = field(
        default=False,
        metadata={
            "help": "Include the dataset name in the example id to make it unique"
        },
    )


@dataclass
class PredictionTargets:
    cea: list[tuple[int, int]]
    cta: list[int]
    cpa: list[tuple[int, int]]


@dataclass
class GPExample(Example[FullTable]):
    kgname: KGName


class DataActor(Actor[DataActorArgs]):
    VERSION = 110

    def __init__(
        self,
        params: DataActorArgs,
        db_actor: (
            DBActor | tuple[DBActor]
        ),  # for __reduce__ & consistent with the super().__init__
    ):
        if isinstance(db_actor, Sequence):
            db_actor = db_actor[0]
        super().__init__(params, [db_actor])
        self.db_actor = db_actor

    @cache(backend=BackendFactory.actor.mem)
    def get_prediction_targets(self, dsquery: str) -> DatasetList[PredictionTargets]:
        """Get the prediction targets for the given dataset query. The prediction targets are
        the columns (single/pair) that we want to predict the type and relation for.
        """
        dataset = DatasetQuery.from_string(dsquery).dataset

        assert dataset.startswith("semtab")
        loc = Datasets().get_dataset(dataset).location

        cea_target_file = loc / "CEA_targets.csv"
        cpa_target_file = loc / "CPA_targets.csv"
        cta_target_file = loc / "CTA_targets.csv"

        cea_lst = M.group_by(serde.csv.deser(cea_target_file), itemgetter(0))
        cpa_lst = M.group_by(serde.csv.deser(cpa_target_file), itemgetter(0))
        cta_lst = M.group_by(serde.csv.deser(cta_target_file), itemgetter(0))

        examples = self.load_dataset(dsquery)
        output = []
        for ex in examples:
            cea = [
                (int(ri) - 1, int(ci))
                for tid, ri, ci in cea_lst.get(ex.table.table.table_id, [])
            ]
            assert all(ri >= 0 for ri, ci in cea)

            tar = PredictionTargets(
                cea=cea,
                cta=[int(ci) for tid, ci in cta_lst.get(ex.table.table.table_id, [])],
                cpa=[
                    (int(ci), int(cj))
                    for tid, ci, cj in cpa_lst.get(ex.table.table.table_id, [])
                ],
            )

            output.append(tar)
        return DatasetList(examples.name, output, examples.provenance)

    @cache(
        backend=BackendFactory.actor.sqlite.serde(
            cls=DatasetList,
            mem_persist=True,
            compression="lz4",
            log_serde_time="DataActor.load_dataset",
        )
    )
    def load_dataset(self, dsquery: str) -> DatasetList[GPExample]:
        parsed_dsquery = DatasetQuery.from_string(dsquery)
        sm_examples = self._load_entire_dataset(parsed_dsquery.dataset)
        ds = parsed_dsquery.select_list(sm_examples)

        if self.params.include_dataset_name:
            for ex in ds:
                assert ex.id == ex.table.table.table_id
                ex.id = f"{ds.name}__{ex.id}"
                ex.table.table.table_id = ex.id

        if parsed_dsquery.postprocessing is not None:
            examples = ds
            for pp in parsed_dsquery.postprocessing:
                if pp == "no-unk-col":
                    fn = ignore_unk_columns
                else:
                    raise NotImplementedError(pp)
                examples = [r for ex in examples if (r := fn(ex)) is not None]
            return DatasetList(ds.name, examples, ds.provenance)
        return ds

    def forward(self, dsquery: str) -> IdentObj[DatasetList[GPExample]]:
        return IdentObj(key=f"{self.key}:{dsquery}", value=self.load_dataset(dsquery))

    @cache(
        backend=BackendFactory.actor.sqlite.pickle(mem_persist=True, compression="lz4")
    )
    def _load_entire_dataset(self, dataset: str) -> list[GPExample]:
        """
        Load dataset either by name or from disk.

        Args:
            dataset: The name of the dataset or the absolute path to the dataset file/folder.

        """
        if dataset.startswith("/"):
            # load dataset from disk
            assert Path(dataset).exists()
            examples = Dataset(Path(dataset)).load()
        else:
            ds = Datasets()
            db = self.get_kgdb(dataset).pydb
            entity_labels = db.entity_labels.cache()
            props = db.props.cache()
            redirections = db.entity_redirections.cache()

            m = re.match(r"(\w+)__v\d+", dataset)
            if m is not None:
                dataset = m.group(1)

            examples = ds.get_dataset(dataset).load()
            examples = ds.fix_redirection(
                examples,
                entity_labels,
                props,
                redirections,
                self.get_kgdb(dataset).kgns,
                skip_unk_ont_ent=self.params.skip_unk_ont_ent,
                skip_no_sm=self.params.skip_no_sm,
            )

        kgname = self.get_kgname(dataset)
        return [
            GPExample(id=ex.id, sms=ex.sms, table=ex.table, kgname=kgname)
            for ex in examples
        ]

    @classmethod
    @lru_cache()
    def get_kgname(cls, dsquery: str) -> KGName:
        dataset = DatasetQuery.from_string(dsquery).dataset
        if (
            dataset.startswith("wt")
            or dataset.startswith("semtab2022_r1")
            or dataset.startswith("semtab2023_r1")
            or dataset.startswith("semtab2024_wikitables")
            or dataset.find("wikidata") != -1
        ):
            return KGName.Wikidata
        if dataset.find("dbpedia") != -1 or dataset.startswith("t2dv2"):
            return KGName.DBpedia
        return KGName.Generic

    def get_kgdb(self, dsquery: str) -> KGDB:
        kgname = self.get_kgname(dsquery)
        return self.db_actor.kgdbs[kgname]

    @cache(backend=BackendFactory.actor.mem)
    def get_sm_helper(self, dsquery: str):
        db = self.get_kgdb(dsquery).pydb
        return SemanticModelHelper(
            db.entity_labels.cache(),
            db.props.cache(),
            get_kgns(self.get_kgname(dsquery)),
        )


# class ExtendedDataActor(DataActor):
#     """An extended data actor that can return datasets generated auto-labeled data actor.
#     Mainly used to manually run cangraph or other actors on the auto-labeled datasets.
#     """

#     VERSION = 100

#     def __init__(
#         self, params: DataActorArgs, autolabel_data_actor: AutoLabeledDataActor
#     ):
#         super(DataActor, self).__init__(params, [autolabel_data_actor])
#         self.db_actor = autolabel_data_actor.db_actor
#         self.autolabel_data_actor = autolabel_data_actor

#     def load_dataset(self, dsquery: str) -> DatasetList[GPExample]:
#         if self.autolabel_data_actor.is_autolabel_dataset(dsquery):
#             ds = self.autolabel_data_actor(dsquery)
#             kgdb = self.get_kgdb(dsquery)
#             kgns = kgdb.kgns
#             entity_label = kgdb.pydb.entity_labels.cache()
#             return DatasetList(
#                 ds.name,
#                 [
#                     from_auto_labeled_table(tbl, kgns, kgdb.kgname, entity_label)
#                     for tbl in ds
#                 ],
#                 ds.provenance,
#             )
#         else:
#             return super().load_dataset(dsquery)


# def from_auto_labeled_table(
#     tbl: AutoLabeledTable,
#     kgns: KnowledgeGraphNamespace,
#     kgname: KGName,
#     entity_label: Mapping[str, str],
# ) -> GPExample:
#     """To convert to the desired format, this method need to convert column types into a
#     semantic model. To support multiple types per column, we create as much classes as
#     needed. For example, column A has two types C1 and C2, then we will have C1 -> A and
#     C2 -> A in the same semantic model.
#     """
#     sm = O.SemanticModel()

#     colids = []
#     for ci, col in enumerate(tbl.table.table.columns):
#         assert col.index == ci
#         cname = col.clean_multiline_name
#         assert cname is not None
#         colids.append(sm.add_node(O.DataNode(ci, cname)))

#     for ci, ctypes in zip(tbl.entity_columns, tbl.entity_column_types):
#         cname = tbl.table.table.get_column_by_index(ci).clean_multiline_name
#         assert cname is not None
#         uid = colids[ci]
#         for ctype in ctypes:
#             assert ctype.id.belong_to(kgns)

#             cid = sm.add_node(
#                 O.ClassNode(
#                     (ctype_uri := kgns.id_to_uri(ctype.id)),
#                     kgns.get_rel_uri(ctype_uri),
#                     readable_label=f"{entity_label[ctype.id]} {ctype.id}",
#                 )
#             )
#             sm.add_edge(O.Edge(cid, uid, str(RDFS.label), "rdfs:label"))

#     return GPExample(
#         id=tbl.table.table.table_id, sms=[sm], table=tbl.table, kgname=kgname
#     )


def ignore_unk_columns(ex: GPExample) -> Optional[GPExample]:
    """Remove unannotated columns in an example."""
    keep_cols = []
    ignore_cols = []

    for col in ex.table.table.columns:
        if all(len(sm.get_semantic_types_of_column(col.index)) > 0 for sm in ex.sms):
            keep_cols.append(col.index)
        else:
            ignore_cols.append(col.index)

    keep_cols = sorted(keep_cols)
    if len(keep_cols) == 0:
        return None

    table = ex.table.keep_columns(keep_cols)
    sms = ex.sms

    if len(ignore_cols) > 0 and ex.sms is not None:
        new_sms = []
        for sm in ex.sms:
            newsm = sm.deep_copy()
            for ci in ignore_cols:
                if newsm.has_data_node(ci):
                    newsm.remove_node(newsm.get_data_node(ci).id)
            new_sms.append(newsm)
        sms = new_sms

    return GPExample(id=ex.id, sms=sms, table=table, kgname=ex.kgname)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal, Optional, Sequence

from gp.actors.data.prelude import GPExample
from gp.actors.el.canreg import CanRegActor
from gp.entity_linking.cangen.common import CanEnt, CanGenMethod, TableCanGenResult

# from gp.entity_linking.cangen.localsearch import LocalSearch, LocalSearchArgs
from gp.entity_linking.cangen.oracle_model import CanGenOracleMethod
from gp.misc.appconfig import AppConfig
from gp.misc.evaluation.el_osin_mixin import EntityLinkingOsinActorMixin
from ream.actor_version import ActorVersion
from ream.actors.base import BaseActor
from ream.cache_helper import Cache, MemBackend
from ream.helper import import_attr
from ream.params_helper import PluginParams
from sm.namespaces.utils import KGName


@dataclass
class CanGenActorArgs(PluginParams):
    # localsearch: Optional[LocalSearchArgs] = field(
    #     default=None,
    #     metadata={
    #         "help": "Search for more candidates by matching the entity's properties with the table's cells"
    #     },
    # )
    soft_limit: int = field(
        default=1000,
        metadata={"help": "The maximum number of candidates per query to be returned."},
    )
    add_gold: Literal["no", "when-not-topk", "always"] = field(
        default="no",
        metadata={
            "help": "add gold entities. options:"
            "- no: never"
            "- when-not-topk: when the gold entity is in candidate entities but not in the top k"
            "- always: add the gold entity when it's missing",
        },
    )


class CanGenActor(
    EntityLinkingOsinActorMixin[CanGenActorArgs], BaseActor[CanGenActorArgs]
):
    """Generate candidate entities for cells in a table."""

    # VERSION = ActorVersion.create(100, [LocalSearch])
    VERSION = 102
    EXP_NAME = "Candidate Generation"
    EXP_VERSION = 3

    def __init__(
        self,
        params: CanGenActorArgs,
        canreg_actor: CanRegActor,
    ):
        super().__init__(params, dep_actors=[canreg_actor])

        self.canreg_actor = canreg_actor
        self.db_actor = canreg_actor.db_actor

    @Cache.cache(
        backend=Cache.sqlite.serde(
            cls=TableCanGenResult, filename="cangen", mem_persist=True
        ),
        cache_key=lambda self, example: example.id,
        disable=lambda self: not AppConfig.get_instance().is_cache_enable,
    )
    def __call__(self, example: GPExample) -> TableCanGenResult:
        ent_col = self.canreg_actor(example)
        res = self.get_method(example.kgname).get_candidates([example], [ent_col])[0]
        return self.fixed_redirections(example.kgname, example.table.table.shape(), res)

    @Cache.flat_cache(
        backend=Cache.sqlite.serde(
            cls=TableCanGenResult, filename="cangen", mem_persist=True
        ),
        cache_key=lambda self, example, verbose=False: example.id,
        disable=lambda self: not AppConfig.get_instance().is_cache_enable,
    )
    def batch_call(
        self, examples: Sequence[GPExample], verbose: bool = False
    ) -> list[TableCanGenResult]:
        if len(examples) == 0:
            return []

        kgname = examples[0].kgname
        assert all(ex.kgname == kgname for ex in examples)

        ent_cols = [self.canreg_actor(ex) for ex in examples]
        return [
            self.fixed_redirections(kgname, examples[ei].table.table.shape(), res)
            for ei, res in enumerate(
                self.get_method(kgname).get_candidates(
                    examples, ent_cols, verbose=verbose
                )
            )
        ]

    def get_candidate_entities(
        self, examples: list[GPExample]
    ) -> list[TableCanGenResult]:
        return self.batch_call(examples)

    def fixed_redirections(
        self, kgname: KGName, shp: tuple[int, int], res: TableCanGenResult
    ) -> TableCanGenResult:
        ent_redirections = self.db_actor.kgdbs[kgname].pydb.entity_redirections.cache()

        if any(id in ent_redirections for id in res.ent_id):
            res = TableCanGenResult.from_matrix(
                res.to_matrix(shp).map(
                    lambda lst: [
                        CanEnt(ent_redirections.get(x.id, x.id), x.score) for x in lst
                    ]
                )
            )
        return res

    def is_oracle_entity_linking(self, kgname: KGName) -> bool:
        return isinstance(
            self.get_method(kgname), CanGenOracleMethod
        ) and self.canreg_actor.is_oracle_entity_linking(kgname)

    @Cache.cache(backend=MemBackend())
    def get_method(self, kgname: KGName) -> CanGenMethod:
        kgdb = self.db_actor.kgdbs[kgname]
        if isinstance(self.params.clspath, str):
            cls = import_attr(self.params.clspath)
        else:
            cls = self.params.clspath

        if isinstance(self.params.clsargs, dict):
            return cls(kgdb, **self.params.clsargs)

        return cls(kgdb, self.params.clsargs)

    # @Cache.cache(
    #     backend=Cache.cls.dir(
    #         cls=DatasetCandidateEntities, mem_persist=True, log_serde_time=True
    #     ),
    #     disable=lambda self: not AppConfig.get_instance().is_cache_enable,
    # )
    # def __call__(self, dsquery: str) -> DatasetCandidateEntities:
    #     examples = self.data_actor(dsquery)
    #     entity_columns = self.canreg_actor(dsquery)

    #     cans = self.method(DatasetQuery.from_string(dsquery).dataset).get_candidates(
    #         dsquery, examples, entity_columns
    #     )

    #     # TODO: this is a special hack for the experiments that we do not want to
    #     # mess up the previous cache as we are limited in time and cannot re-compute the
    #     # ranking

    #     # we want to add the gold entity to the candidates if it's not there.
    #     if self.params.soft_limit == 99:
    #         # add if it's not in the top 100
    #         self.logger.warning(
    #             "We are adding gold entities to the candidates. You should know what you are doing!"
    #         )
    #         gold_ents: dict[str, dict[int, dict[int, list[CandidateEntity]]]] = (
    #             defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
    #         )
    #         db = self.data_actor.get_kgdb(dsquery).pydb

    #         entity_metadata = db.entity_metadata.cache()
    #         entity_pagerank = db.entity_pagerank.cache()

    #         for ex in examples:
    #             table_id = ex.table.table.table_id
    #             for ri, ci, links in ex.table.links.enumerate_flat_iter():
    #                 if len(links) > 0:
    #                     for link in links:
    #                         for entity_id in link.entities:
    #                             entity_id = str(entity_id)
    #                             ent = entity_metadata[entity_id]
    #                             gold_ents[table_id][ci][ri].append(
    #                                 CandidateEntity(
    #                                     id=ent.id,
    #                                     label=str(ent.label),
    #                                     description=str(ent.description),
    #                                     aliases=sorted(ent.aliases.get_all()),
    #                                     popularity=entity_pagerank[ent.id],
    #                                     score=0.0,
    #                                 )
    #                             )
    #         cans = cans.top_k_candidates(self.params.soft_limit, gold_ents)

    #     if self.params.localsearch is not None:
    #         cans = LocalSearch(
    #             self.params.localsearch, self.data_actor.get_kgdb(dsquery)
    #         ).augment_candidates(examples, cans)
    #     return cans



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from gp.actors.data.prelude import GPExample
from gp.actors.el.cangen import CanGenActor
from gp.entity_linking.cangen.common import TableCanGenResult
from gp.entity_linking.canrank.common import CanRankMethod, TableCanGenUpdateScores
from gp.misc.appconfig import AppConfig
from gp.misc.evaluation.el_osin_mixin import EntityLinkingOsinActorMixin
from ream.actors.base import BaseActor
from ream.cache_helper import Cache, MemBackend
from ream.helper import import_attr
from ream.params_helper import PluginParams


@dataclass
class CanRankActorArgs(PluginParams):
    pass


class CanRankActor(
    EntityLinkingOsinActorMixin[CanRankActorArgs], BaseActor[CanRankActorArgs]
):
    """Rank candidate entities for cells in a table."""

    VERSION = 103

    EXP_NAME = "Candidate Ranking"
    EXP_VERSION = 3

    def __init__(
        self,
        params: CanRankActorArgs,
        cangen_actor: CanGenActor,
    ):
        super().__init__(params, dep_actors=[cangen_actor])

        self.cangen_actor = cangen_actor
        self.db_actor = cangen_actor.db_actor

    def __call__(self, example: GPExample) -> TableCanGenResult:
        cans = self.cangen_actor(example)
        cans = cans.shallow_clone()
        cans.ent_score = self.impl_call(example).value
        return cans

    def batch_call(
        self, examples: Sequence[GPExample], verbose: bool = False
    ) -> list[TableCanGenResult]:
        ex_cans = self.cangen_actor.batch_call(examples)
        ex_cans_scores = self.impl_batch_call(examples, verbose)
        output = []
        for ei in range(len(examples)):
            cans = ex_cans[ei].shallow_clone()
            cans.ent_score = ex_cans_scores[ei].value
            output.append(cans)
        return output

    @Cache.cache(
        backend=Cache.sqlite.serde(
            cls=TableCanGenUpdateScores, filename="canrank", mem_persist=True
        ),
        cache_key=lambda self, example: example.id,
        disable=lambda self: not AppConfig.get_instance().is_cache_enable,
    )
    def impl_call(self, example: GPExample) -> TableCanGenUpdateScores:
        return self.get_method().batch_rank_candidates([example])[0]

    @Cache.flat_cache(
        backend=Cache.sqlite.serde(
            cls=TableCanGenUpdateScores, filename="canrank", mem_persist=True
        ),
        cache_key=lambda self, example, verbose=False: example.id,
        disable=lambda self: not AppConfig.get_instance().is_cache_enable,
    )
    def impl_batch_call(
        self, examples: Sequence[GPExample], verbose: bool = False
    ) -> list[TableCanGenUpdateScores]:
        return self.get_method().batch_rank_candidates(examples, verbose)

    @Cache.cache(backend=MemBackend())
    def get_method(self) -> CanRankMethod:
        if isinstance(self.params.clspath, str):
            cls = import_attr(self.params.clspath)
        else:
            cls = self.params.clspath

        if isinstance(self.params.clsargs, dict):
            return cls(**self.params.clsargs, cangen_actor=self.cangen_actor)

        return cls(self.params.clsargs, cangen_actor=self.cangen_actor)

    def get_candidate_entities(
        self, examples: list[GPExample]
    ) -> list[TableCanGenResult]:
        return self.batch_call(examples)



from __future__ import annotations

from dataclasses import dataclass, field

from gramsplus.core.algorithms import CanGraphExtractorConfig
from gramsplus.semanticmodeling.literal_matcher import LiteralMatcherConfig


@dataclass
class CanGraphExtractorCfg:
    literal_matcher: LiteralMatcherConfig = field(
        default_factory=LiteralMatcherConfig,
        metadata={"help": "Configuration for the literal matcher"},
    )
    n_hop: int = field(
        default=1,
        metadata={
            "help": "The number of hops to discover relationships between entities"
        },
    )
    ignored_columns: list[int] = field(
        default_factory=list,
        metadata={
            "help": "The columns to ignore and not used during **literal** matching"
        },
    )
    ignored_props: set[str] = field(
        default_factory=set,
        metadata={
            "help": "The properties to ignore and not used during **literal** matching"
        },
    )
    allow_same_ent_search: bool = field(
        default=False,
        metadata={
            "help": "whether we try to discover relationships between the same entity in same row but different columns"
        },
    )
    allow_ent_matching: bool = field(
        default=True,
        metadata={
            "help": "whether to match relationships between entities (i.e., if false only do literal matching)"
        },
    )
    use_context: bool = field(
        default=False,
        metadata={"help": "whether we use context entities in the search"},
    )
    add_missing_property: bool = field(
        default=True,
        metadata={
            "help": "whether to add a new node to represent a statement's property value if the statement has both qualifier and property pointing to the same target node to give the qualifier a chance to be selected."
        },
    )
    run_subproperty_inference: bool = field(
        default=False,
        metadata={
            "help": "whether to run subproperty inference to complete missing links"
        },
    )
    run_transitive_inference: bool = field(
        default=True,
        metadata={
            "help": "whether to run transitive inference to complete missing links"
        },
    )
    deterministic_order: bool = field(
        default=False,
        metadata={
            "help": "whether to sort the matched results so that the order is always deterministic"
        },
    )
    correct_entity_threshold: float = field(
        default=0.8,
        metadata={
            "help": "the threshold that a candidate entity is considered correct for contradicted information detection"
        },
    )
    validate: bool = field(
        default=False,
        metadata={
            "help": "whether to validate the result candidate graph. This is useful for debugging"
        },
    )

    def to_rust(self) -> CanGraphExtractorConfig:
        return CanGraphExtractorConfig(
            self.literal_matcher.to_rust(),
            self.ignored_columns,
            self.ignored_props,
            self.allow_same_ent_search,
            self.allow_ent_matching,
            self.use_context,
            self.add_missing_property,
            self.run_subproperty_inference,
            self.run_transitive_inference,
            self.deterministic_order,
            self.correct_entity_threshold,
            self.validate,
            self.n_hop,
        )



from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass
from typing import Mapping, Sequence, TypeAlias

from gp.semanticmodeling.postprocessing.cgraph import CGEdgeTriple


@dataclass
class PostprocessingResult:
    edges: Sequence[CGEdgeTriple]
    nodes: Mapping[int, str]


EdgeProb: TypeAlias = dict[CGEdgeTriple, float]
# mapping from column to each type and its probability
NodeProb: TypeAlias = dict[int, dict[str, float]]


class PostprocessingFn:
    @abstractmethod
    def __call__(
        self,
        node_probs: NodeProb,
        edge_probs: EdgeProb,
    ) -> PostprocessingResult: ...



from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property, lru_cache
from linecache import cache
from typing import Optional

from gp.actors.data.prelude import KGDB, GPExample, KGDBArgs
from gp.actors.el.canrank import CanRankActor
from gp.misc.appconfig import AppConfig
from gp.misc.conversions import to_rust_table
from gp.semanticmodeling import text_parser
from gp.semanticmodeling.cangraph_cfg import CanGraphExtractorCfg
from gp.semanticmodeling.text_parser import TextParser, TextParserConfigs
from gramsplus.core.algorithms import (
    CanGraphExtractedResult,
    extract_cangraph,
    par_extract_cangraphs,
)
from gramsplus.core.models import TableCells
from ream.prelude import BaseActor, Cache
from sm.dataset import Example, FullTable
from sm.misc.ray_helper import enhance_error_info, ray_map, ray_put
from tqdm import tqdm


@dataclass
class CanGraphActorArgs:
    topk: Optional[int] = field(
        default=100,
        metadata={
            "help": "keeping maximum top k candidates for each cell",
        },
    )
    text_parser: TextParserConfigs = field(
        default_factory=TextParserConfigs,
        metadata={"help": "Configuration for the text parser"},
    )
    cangraph_extractor: CanGraphExtractorCfg = field(
        default_factory=CanGraphExtractorCfg,
        metadata={"help": "Configuration for the candidate graph extractor"},
    )


class CanGraphActor(BaseActor[CanGraphActorArgs]):
    VERSION = 117

    def __init__(self, params: CanGraphActorArgs, canrank_actor: CanRankActor):
        super().__init__(params, dep_actors=[canrank_actor])

        self.canrank_actor = canrank_actor
        self.db_actor = canrank_actor.db_actor

    @cached_property
    def text_parser(self) -> TextParser:
        return TextParser(self.params.text_parser)

    @cached_property
    def ru_cangraph_extractor(self):
        return self.params.cangraph_extractor.to_rust()

    @Cache.cache(
        backend=Cache.sqlite.pickle(
            filename="cangraph", mem_persist=True, compression="lz4"
        ),
        cache_key=lambda self, example, parallel=True: example.id,
        disable=lambda self: not AppConfig.get_instance().is_cache_enable,
    )
    def __call__(self, ex: GPExample, parallel: bool = True) -> CanGraphExtractedResult:
        cans = self.get_candidate_entities([ex])[0]
        nrows, ncols = ex.table.table.shape()
        text_parser = self.text_parser

        return extract_cangraph(
            to_rust_table(ex, cans),
            TableCells(
                [
                    [
                        text_parser.parse(ex.table.table[ri, ci]).to_rust()
                        for ci in range(ncols)
                    ]
                    for ri in range(nrows)
                ]
            ),
            self.db_actor.kgdbs[ex.kgname].rudb,
            self.ru_cangraph_extractor,
            None,
            parallel=parallel,
        )

    @Cache.flat_cache(
        backend=Cache.sqlite.pickle(
            filename="cangraph", mem_persist=True, compression="lz4"
        ),
        cache_key=lambda self, example, verbose=False: example.id,
        disable=lambda self: not AppConfig.get_instance().is_cache_enable,
    )
    def batch_call(
        self, exs: list[GPExample], verbose: bool = False
    ) -> list[CanGraphExtractedResult]:
        if len(exs) == 0:
            return []

        assert all(
            ex.kgname == exs[0].kgname for ex in exs
        ), "Don't support multiple KGs in the same function call"

        ex_cans = self.get_candidate_entities(exs)
        text_parser = self.text_parser

        tables = []
        table_cells = []

        for ei, ex in tqdm(enumerate(exs), disable=not verbose):
            nrows, ncols = ex.table.table.shape()
            tbl = to_rust_table(ex, ex_cans[ei])
            cells = TableCells(
                [
                    [
                        text_parser.parse(ex.table.table[ri, ci]).to_rust()
                        for ci in range(ncols)
                    ]
                    for ri in range(nrows)
                ]
            )

            tables.append(tbl)
            table_cells.append(cells)

        return par_extract_cangraphs(
            tables,
            table_cells,
            self.db_actor.kgdbs[exs[0].kgname].rudb,
            self.ru_cangraph_extractor,
            None,
            verbose=verbose,
        )

    def get_candidate_entities(self, exs: list[GPExample]):
        ex_cans = self.canrank_actor.batch_call(exs)
        if self.params.topk is not None:
            return [cans.top_k(self.params.topk) for cans in ex_cans]
        return ex_cans



from __future__ import annotations

from collections import defaultdict
from collections.abc import Mapping
from dataclasses import dataclass
from operator import itemgetter
from typing import Optional, Sequence

from gp.misc.evaluation.sm_wikidata import SemanticModelHelper
from gp.misc.itemdistance import get_hierarchy_scoring_fn
from graph.retworkx.api import digraph_all_simple_paths
from kgdata.models import Ontology
from loguru import logger
from sm.dataset import Example, FullTable
from sm.evaluation.cpa_cta_metrics import _cta_real, _get_cta
from sm.evaluation.prelude import (
    CTAEvalOutput,
    PrecisionRecallF1,
    _cpa_transformation,
    cta,
    sm_metrics,
)
from sm.evaluation.sm_metrics import PermutationExplosion
from sm.misc.funcs import assert_isinstance
from sm.outputs.semantic_model import ClassNode, DataNode
from sm.outputs.semantic_model import Edge as SMEdge
from sm.outputs.semantic_model import LiteralNode, SemanticModel


@dataclass
class WrappedSemanticModel:
    sm: SemanticModel
    is_normalized: bool = False
    is_cpa_transformed: bool = False


@dataclass
class CPAEvalRes:
    result: PrecisionRecallF1
    gold_sm: SemanticModel
    cpa_gold_sm: SemanticModel
    cpa_pred_sm: SemanticModel


@dataclass
class CTAEvalRes:
    result: CTAEvalOutput
    # mapping from column index (converted to string) to class id
    cta_gold: dict[str, str]
    # mapping from column index (converted to string) to class id
    cta_pred: dict[str, str]


class Evaluator:
    def __init__(
        self,
        ontology: Ontology,
        entity_labels: Optional[Mapping[str, str]] = None,
    ):
        self.sm_helper = SemanticModelHelper(
            entity_labels, ontology.props, ontology.kgns
        )
        self.kgns = self.sm_helper.kgns
        self.class_scoring_fn = get_hierarchy_scoring_fn(ontology.classes, self.kgns)
        self.prop_scoring_fn = get_hierarchy_scoring_fn(ontology.props, self.kgns)

    def avg_cpa(
        self,
        examples: Sequence[Example[FullTable]],
        sms: Sequence[SemanticModel],
        extra_cols: tuple[str, str] = ("", ""),
    ) -> list[CPAEvalRes]:
        cpas: list[CPAEvalRes] = []
        for ex, pred_sm in zip(examples, sms):
            cpas.append(self.cpa(ex, pred_sm))

        avg_cpa = PrecisionRecallF1.avg([cpa.result for cpa in cpas])
        logger.info(
            "for copying...\n{}\tcpa-p\tcpa-r\tcpa-f1\n{}",
            (extra_cols[0] + "\t") if extra_cols[0] else "",
            ",".join(
                ([extra_cols[1]] if extra_cols[1] else [])
                + [
                    "%.2f" % (round(float(x) * 100, 2))
                    for x in [avg_cpa.precision, avg_cpa.recall, avg_cpa.f1]
                ],
            ),
        )
        # fmt: on
        return cpas

    def avg_cta(
        self,
        examples: Sequence[Example[FullTable]],
        sms: Sequence[
            SemanticModel | Mapping[str, str] | Mapping[str, dict[str, float]]
        ],
        extra_cols: tuple[str, str] = ("", ""),
    ) -> list[CTAEvalOutput]:
        ctas: list[CTAEvalOutput] = []
        for ex, pred_sm in zip(examples, sms):
            ctas.append(self.cta(ex, pred_sm).result)
        avg_cta = PrecisionRecallF1.avg(ctas)
        logger.info(
            "for copying...\n{}\tcta-p\tcta-r\tcta-f1\n{}",
            (extra_cols[0] + "\t") if extra_cols[0] else "",
            ",".join(
                ([extra_cols[1]] if extra_cols[1] else [])
                + [
                    "%.2f" % (round(float(x) * 100, 2))
                    for x in [avg_cta.precision, avg_cta.recall, avg_cta.f1]
                ],
            ),
        )
        # fmt: on
        return ctas

    def cpa(
        self,
        example: Example[FullTable],
        sm: SemanticModel,
    ) -> CPAEvalRes:
        """Calculate the CPA score. The code is borrowed from: sm.evaluation.cpa_cta_metrics.cpa to adapt with this class API that uses WrappedSemanticModel"""
        gold_sms = self.get_example_gold_sms(example)
        cpa_pred_sm = self.convert_sm_for_cpa(self.norm_sm(sm))
        cpa_gold_sms = [self.convert_sm_for_cpa(gold_sm) for gold_sm in gold_sms]
        return self._cpa_real(
            cpa_pred_sm,
            cpa_gold_sms,
            gold_sms,
        )

    def _cpa_real(
        self,
        cpa_pred_sm: WrappedSemanticModel,
        cpa_gold_sms: list[WrappedSemanticModel],
        gold_sms: list[WrappedSemanticModel],
    ) -> CPAEvalRes:
        """Calculate the CPA score. The code is borrowed from: sm.evaluation.cpa_cta_metrics.cpa to adapt with this class API that uses WrappedSemanticModel"""
        assert cpa_pred_sm.is_normalized and cpa_pred_sm.is_cpa_transformed
        assert all(
            cpa_gold_sm.is_normalized and cpa_gold_sm.is_cpa_transformed
            for cpa_gold_sm in cpa_gold_sms
        )

        output = None
        best_cpa_gold_sm = None
        best_gold_sm = None

        for i, cpa_gold_sm in enumerate(cpa_gold_sms):
            try:
                res = sm_metrics.precision_recall_f1(
                    gold_sm=cpa_gold_sm.sm,
                    pred_sm=cpa_pred_sm.sm,
                    scoring_fn=self.prop_scoring_fn,
                    # debug_dir="/tmp",
                    hard_permutation_threshold=100000,
                )
            except PermutationExplosion:
                res = sm_metrics.precision_recall_f1(
                    gold_sm=self.uniquify_statement(cpa_gold_sm.sm),
                    pred_sm=self.uniquify_statement(cpa_pred_sm.sm),
                    scoring_fn=self.prop_scoring_fn,
                    debug_dir="/tmp",
                    hard_permutation_threshold=100000,
                )

            if output is None or res.f1 > output.f1:
                output = res
                best_cpa_gold_sm = cpa_gold_sm
                best_gold_sm = gold_sms[i]

        assert (
            output is not None
            and best_cpa_gold_sm is not None
            and best_gold_sm is not None
        )

        return CPAEvalRes(
            result=output,
            gold_sm=best_gold_sm.sm,
            cpa_gold_sm=best_cpa_gold_sm.sm,
            cpa_pred_sm=cpa_pred_sm.sm,
        )

    def cta(
        self,
        example: Example[FullTable],
        pred_sm: SemanticModel | Mapping[str, str] | Mapping[str, dict[str, float]],
    ) -> CTAEvalRes:
        if isinstance(pred_sm, Mapping):
            pred_cta = {}
            for k, v in pred_sm.items():
                if isinstance(v, dict):
                    curi = max(v.items(), key=itemgetter(1))[0]
                else:
                    curi = v

                if not self.kgns.is_uri(curi):
                    assert self.kgns.is_id(curi)
                    curi = self.kgns.id_to_uri(curi)
                pred_cta[k] = curi

            gold_cta, cta_output = max(
                [
                    (
                        (gold_cta := _get_cta(gold_sm.sm, self.sm_helper.ID_PROPS)),
                        _cta_real(
                            gold_cta,
                            pred_cta,
                            self.class_scoring_fn,
                        ),
                    )
                    for gold_sm in self.get_example_gold_sms(example)
                ],
                key=lambda x: x[1].f1,
            )
        else:
            gold_sms = self.get_example_gold_sms(example)
            wrapped_pred_sm = self.norm_sm(pred_sm)
            gold_sm, cta_output = max(
                [
                    (
                        gold_sm,
                        cta(
                            gold_sm.sm,
                            wrapped_pred_sm.sm,
                            self.sm_helper.ID_PROPS,
                            self.class_scoring_fn,
                        ),
                    )
                    for gold_sm in gold_sms
                ],
                key=lambda x: x[1].f1,
            )

            pred_cta = _get_cta(wrapped_pred_sm.sm, self.sm_helper.ID_PROPS)
            gold_cta = _get_cta(gold_sm.sm, self.sm_helper.ID_PROPS)
        return CTAEvalRes(cta_output, gold_cta, pred_cta)

    # @CacheMethod.cache(CacheMethod.single_object_arg)
    def get_example_gold_sms(
        self, example: Example[FullTable]
    ) -> list[WrappedSemanticModel]:
        return self.get_equiv_sms(example.sms)

    # @CacheMethod.cache(CacheMethod.single_object_arg)
    def norm_sm(self, sm: SemanticModel) -> WrappedSemanticModel:
        return WrappedSemanticModel(self.sm_helper.norm_sm(sm), is_normalized=True)

    # @CacheMethod.cache(CacheMethod.single_object_arg)
    def convert_sm_for_cpa(
        self, wrapped_sm: WrappedSemanticModel
    ) -> WrappedSemanticModel:
        """Convert a semantic model to another model for evaluating the CPA task:
        - SemModelTransformation.replace_class_nodes_by_subject_columns(sm, id_props)
        - SemModelTransformation.remove_isolated_nodes(sm)
        """
        assert wrapped_sm.is_normalized
        if wrapped_sm.is_cpa_transformed:
            return wrapped_sm

        sm = wrapped_sm.sm
        cpa_sm = sm.deep_copy()
        _cpa_transformation(cpa_sm, self.sm_helper.ID_PROPS)

        return WrappedSemanticModel(cpa_sm, is_normalized=True, is_cpa_transformed=True)

    def get_equiv_sms(self, sms: list[SemanticModel]) -> list[WrappedSemanticModel]:
        return [
            WrappedSemanticModel(equiv_sm, is_normalized=True)
            for sm in sms
            for equiv_sm in self.sm_helper.gen_equivalent_sm(
                sm, strict=False, incorrect_invertible_props={"P571", "P582"}
            )
        ]

    def uniquify_statement(self, sm: SemanticModel) -> SemanticModel:
        """For handle permutation issue when evaluating precision/recall/f1"""
        newsm = sm.deep_copy()
        for node in newsm.iter_nodes():
            if isinstance(node, ClassNode):
                if node.abs_uri == self.sm_helper.kgns.statement_uri:
                    inedges = newsm.in_edges(node.id)
                    outedges = newsm.out_edges(node.id)

                    if len(inedges) == 1 and len(outedges) <= 1:
                        postfix = "_" + inedges[0].rel_uri
                        node.abs_uri = self.sm_helper.kgns.statement_uri + postfix
                        node.rel_uri = self.sm_helper.kgns.statement_uri + postfix
        return newsm

    def get_cpa_between_columns(
        self, sms: list[WrappedSemanticModel]
    ) -> set[tuple[int, int, str, str]]:
        """Get relationships between columns in a form of (source column, target column, property, qualifier).

        Note: This function will NOT generate relationships to/from entity or literal nodes.
        """
        rels: set[tuple[int, int, str, str]] = set()
        for sm in sms:
            cpa_sm = self.convert_sm_for_cpa(sm)
            for snode in cpa_sm.sm.iter_nodes():
                if not (
                    isinstance(snode, ClassNode)
                    and snode.abs_uri == self.kgns.statement_uri
                ):
                    continue

                (usedge,) = cpa_sm.sm.in_edges(snode.id)
                u = cpa_sm.sm.get_node(usedge.source)
                if not isinstance(u, DataNode):
                    continue
                sourcecol = u.col_index

                for svedge in cpa_sm.sm.out_edges(snode.id):
                    v = cpa_sm.sm.get_node(svedge.target)
                    if not isinstance(v, DataNode):
                        continue
                    targetcol = v.col_index
                    rels.add(
                        (
                            sourcecol,
                            targetcol,
                            self.kgns.uri_to_id(usedge.abs_uri),
                            self.kgns.uri_to_id(svedge.abs_uri),
                        )
                    )
        return rels

    def remove_incorrect_relationships(
        self, pred_sm: SemanticModel, gold_sms: list[WrappedSemanticModel]
    ) -> WrappedSemanticModel:
        """Remove easily detected incorrect relationships from the predicted semantic model to measure
        the maximum recall of the predicted semantic model. This is often used to measure the quality of
        candidate graphs.

        The returned result is a normalized CPA model.
        """
        gold_rels: set[tuple[int, int, str, str]] = self.get_cpa_between_columns(
            gold_sms
        )

        # remove relationships that aren't in the ground truth.
        newsm = self.convert_sm_for_cpa(self.norm_sm(pred_sm))
        for snode in newsm.sm.nodes():
            if (
                isinstance(snode, ClassNode)
                and snode.abs_uri == self.kgns.statement_uri
            ):
                (usedge,) = newsm.sm.in_edges(snode.id)
                u = newsm.sm.get_node(usedge.source)
                assert isinstance(u, DataNode)
                sourcecol = u.col_index

                for svedge in newsm.sm.out_edges(snode.id):
                    v = newsm.sm.get_node(svedge.target)
                    assert isinstance(v, DataNode)
                    targetcol = v.col_index
                    if (
                        sourcecol,
                        targetcol,
                        self.kgns.uri_to_id(usedge.abs_uri),
                        self.kgns.uri_to_id(svedge.abs_uri),
                    ) not in gold_rels:
                        newsm.sm.remove_edge(svedge.id)

                if newsm.sm.out_degree(snode.id) == 0:
                    newsm.sm.remove_node(snode.id)

        # remove duplicate edges
        dup_edges = defaultdict(list)
        for snode in newsm.sm.nodes():
            if not (
                isinstance(snode, ClassNode)
                and snode.abs_uri == self.kgns.statement_uri
                and newsm.sm.out_degree(snode.id) == 1
            ):
                continue

            (usedge,) = newsm.sm.in_edges(snode.id)
            u = newsm.sm.get_node(usedge.source)
            assert isinstance(u, DataNode)
            sourcecol = u.col_index

            (svedge,) = newsm.sm.out_edges(snode.id)
            v = newsm.sm.get_node(svedge.target)
            assert isinstance(v, DataNode)
            targetcol = v.col_index

            key = (
                sourcecol,
                targetcol,
                usedge.abs_uri,
                svedge.abs_uri,
            )
            dup_edges[key].append(snode.id)

        for key, snode_ids in dup_edges.items():
            if len(snode_ids) > 1:
                for snode_id in snode_ids[1:]:
                    newsm.sm.remove_node(snode_id)

        return newsm

    def keep_targets(
        self, sm: SemanticModel, cpa: Sequence[tuple[int, int]], cta: Sequence[int]
    ) -> SemanticModel:
        wsm = self.norm_sm(sm)
        wsm = WrappedSemanticModel(
            wsm.sm.copy(),
            is_normalized=wsm.is_normalized,
            is_cpa_transformed=wsm.is_cpa_transformed,
        )
        # first of all, we need to validate the targets, if a column is not an entity column, it cannot participate in the CPA
        entcols = set(cta)
        for ci, cj in cpa:
            if ci not in entcols:
                raise ValueError(
                    f"Column {ci} is not an entity column but it has outgoing relationships"
                )

        # remove literal nodes such as entity nodes
        for v in list(wsm.sm.iter_nodes()):
            if isinstance(v, LiteralNode):
                wsm.sm.remove_node(v.id)
        for v in list(wsm.sm.iter_nodes()):
            if (
                isinstance(v, ClassNode)
                and v.abs_uri == self.kgns.statement_uri
                and (wsm.sm.in_degree(v.id) == 0 or wsm.sm.out_degree(v.id) == 0)
            ):
                wsm.sm.remove_node(v.id)

        # remove columns that are tagged as entity column but not in the target list first -- don't need to remove standalone data nodes
        for v in wsm.sm.iter_data_nodes():
            if (
                not wsm.sm.is_entity_column(v.col_index, self.sm_helper.ID_PROPS)
                or v.col_index in cta
            ):
                continue

            # find the class node <u> that is connected to this column and remove it
            (uvedge,) = wsm.sm.in_edges(v.id)
            u = wsm.sm.get_node(uvedge.source)
            assert isinstance(u, ClassNode) and u.abs_uri != self.kgns.statement_uri

            # now to remove <u>, we need to remove the statement nodes as well
            for usedge in wsm.sm.out_edges(u.id):
                s = wsm.sm.get_node(usedge.target)
                if isinstance(s, ClassNode):
                    assert (
                        s.abs_uri == self.kgns.statement_uri
                        and wsm.sm.in_degree(usedge.target) == 1
                    )
                    wsm.sm.remove_node(usedge.target)
                else:
                    assert isinstance(s, DataNode) and s.col_index == v.col_index
            # we need to rewire the edge to the column node before removing <u>
            for suedge in wsm.sm.in_edges(u.id):
                wsm.sm.add_edge(
                    SMEdge(
                        source=suedge.source,
                        target=v.id,
                        abs_uri=suedge.abs_uri,
                        rel_uri=suedge.rel_uri,
                        approximation=suedge.approximation,
                        readable_label=suedge.readable_label,
                    )
                )
            wsm.sm.remove_node(u.id)

        # remove relationships that are not in the cpa
        cpa_wsm = self.convert_sm_for_cpa(wsm)
        relcols = set(cpa)
        remove_rels: set[tuple[int, int]] = set()

        for s in cpa_wsm.sm.iter_nodes():
            if isinstance(s, ClassNode):
                assert (
                    s.abs_uri == self.kgns.statement_uri
                ), "The cpa model only contains statement and data nodes"
                (uvedge,) = cpa_wsm.sm.in_edges(s.id)
                for svedge in cpa_wsm.sm.out_edges(s.id):
                    u = assert_isinstance(cpa_wsm.sm.get_node(uvedge.source), DataNode)
                    v = assert_isinstance(cpa_wsm.sm.get_node(svedge.target), DataNode)
                    if (u.col_index, v.col_index) not in relcols:
                        remove_rels.add(
                            (
                                u.col_index,
                                v.col_index,
                            )
                        )

        for ci, cj in remove_rels:
            # we need to find the correct source & target nodes to remove the relationship
            if wsm.sm.is_entity_column(ci, self.sm_helper.ID_PROPS):
                (inedge,) = wsm.sm.in_edges(wsm.sm.get_data_node(ci).id)
                u = wsm.sm.get_node(inedge.source)
                # class node connects with entity columns must not be the statement node (normalization exclude id props)
                assert isinstance(u, ClassNode) and u.abs_uri != self.kgns.statement_uri
            else:
                u = wsm.sm.get_data_node(ci)

            if wsm.sm.is_entity_column(cj, self.sm_helper.ID_PROPS):
                (inedge,) = wsm.sm.in_edges(wsm.sm.get_data_node(cj).id)
                v = wsm.sm.get_node(inedge.source)
                # class node connects with entity columns must not be the statement node (normalization exclude id props)
                assert isinstance(v, ClassNode) and v.abs_uri != self.kgns.statement_uri
            else:
                v = wsm.sm.get_data_node(cj)

            paths = digraph_all_simple_paths(wsm.sm, u.id, v.id, cutoff=2)
            for path in paths:
                inedge, outedge = path
                wsm.sm.remove_edge(inedge.id)
                wsm.sm.remove_edge(outedge.id)

                s = wsm.sm.get_node(inedge.target)
                assert isinstance(s, ClassNode) and s.abs_uri == self.kgns.statement_uri
                if wsm.sm.degree(inedge.target) == 0:
                    wsm.sm.remove_node(inedge.target)

        return wsm.sm



from __future__ import annotations

import re
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Mapping,
    Optional,
    Sequence,
    TypeAlias,
    TypedDict,
    Union,
)

from gp.misc.evaluation.sm_wikidata import SemanticModelHelper
from gp_core.models import GramsDB, LocalGramsDB, RemoteGramsDB
from kgdata.db import GenericDB
from kgdata.models import Ontology
from kgdata.wikidata.db import WikidataDB
from libactor.actor import Actor
from libactor.cache import IdentObj
from sm.misc.prelude import import_attr
from sm.misc.ray_helper import get_instance
from sm.namespaces.namespace import KnowledgeGraphNamespace
from sm.namespaces.utils import KGName, get_kgns


@dataclass
class KGDBArgs:
    name: KGName
    version: str
    datadir: Path
    clspath: Optional[str] = None
    entity_url: Optional[str] = None
    entity_metadata_url: Optional[str] = None
    entity_batch_size: int = 64
    entity_metadata_batch_size: int = 128

    def to_dict(self):
        return {"name": self.name, "version": self.version}

    def get_entity_urls(self) -> list[str]:
        return self.get_urls(self.entity_url)

    def get_entity_metadata_urls(self) -> list[str]:
        return self.get_urls(self.entity_metadata_url)

    def get_urls(self, pattern: Optional[str]) -> list[str]:
        if pattern is None:
            return []
        ((begin, end),) = re.findall(r"(\d+)-(\d+)", pattern)
        template = pattern.replace(f"{begin}-{end}", "{}")
        return [template.format(i) for i in range(int(begin), int(end))]

    def get_key(self):
        return f"{self.name}:{self.version}:{self.datadir}"


@dataclass
class DBActorArgs:
    kgdbs: Sequence[KGDBArgs]

    def __post_init__(self):
        self.kgdbs = sorted(self.kgdbs, key=lambda db: db.name)
        assert len(set(db.name for db in self.kgdbs)) == len(
            self.kgdbs
        ), "Should not have duplicated KGs"

    def to_dict(self):
        return [{"name": kgdb.name, "version": kgdb.version} for kgdb in self.kgdbs]


class DBActor(Actor[DBActorArgs]):
    VERSION = 100

    def __init__(self, params: DBActorArgs):
        super().__init__(params, [])
        self.kgdbs: Mapping[KGName, KGDB] = {
            kgdb.name: KGDB.get_instance(kgdb) for kgdb in params.kgdbs
        }


PyKGDB: TypeAlias = Union[GenericDB, WikidataDB]


@dataclass
class KGDB:
    args: KGDBArgs

    @cached_property
    def kgns(self) -> KnowledgeGraphNamespace:
        return get_kgns(self.args.name)

    @property
    def kgname(self):
        return self.args.name

    @cached_property
    def pydb(self) -> PyKGDB:
        if self.args.name == KGName.Wikidata:
            return WikidataDB(self.args.datadir)
        if self.args.name == KGName.DBpedia:
            return GenericDB(self.args.datadir)
        if self.args.name == KGName.Generic:
            if self.args.clspath is not None:
                cls = import_attr(self.args.clspath)
            else:
                cls = GenericDB
            return cls(self.args.datadir)
        raise NotImplementedError(self.args.name)

    @cached_property
    def rudb(self) -> GramsDB:
        if self.args.entity_url is None:
            return LocalGramsDB(str(self.args.datadir))
        else:
            return RemoteGramsDB(
                str(self.args.datadir),
                self.args.get_entity_urls(),
                self.args.get_entity_metadata_urls(),
                self.args.entity_batch_size,
                self.args.entity_metadata_batch_size,
            )

    @cached_property
    def sm_helper(self):
        return SemanticModelHelper(
            self.pydb.entity_labels.cache(),
            self.pydb.props.cache(),
            self.kgns,
        )

    @cached_property
    def ontology(self) -> IdentObj[Ontology]:
        return IdentObj(
            key=f"ontology:{self.args.name}:{self.args.version}",
            value=Ontology(
                kgname=self.kgname,
                kgns=self.kgns,
                classes=self.pydb.classes.cache(),
                props=self.pydb.props.cache(),
            ),
        )

    @staticmethod
    def get_instance(db: KGDB | KGDBArgs) -> KGDB:
        if isinstance(db, KGDB):
            return db
        return get_instance(
            lambda: KGDB(db), f"kgdb:{db.name}:{db.version}:{db.datadir}"
        )



from typing import Optional

class LiteralMatcherConfig:
    string: str
    quantity: str
    globecoordinate: str
    time: str
    monolingual_text: str
    entity: str

    def __init__(
        self,
        string: str,
        quantity: str,
        globecoordinate: str,
        time: str,
        monolingual_text: str,
        entity: str,
    ): ...
    def save(self, outfile: str) -> None: ...

class LiteralMatcher:
    def __init__(self, cfg: LiteralMatcherConfig): ...

class ParsedTextRepr:
    origin: str
    normed_string: str
    number: Optional[ParsedNumberRepr]
    datetime: Optional[ParsedDatetimeRepr]

    def __init__(
        self,
        origin: str,
        normed_string: str,
        number: Optional[ParsedNumberRepr],
        datetime: Optional[ParsedDatetimeRepr],
    ): ...

class ParsedNumberRepr:
    number: float
    number_string: str
    is_integer: bool
    unit: Optional[str]
    prob: Optional[float]

    def __init__(
        self,
        number: float,
        number_string: str,
        is_integer: bool,
        unit: Optional[str],
        prob: Optional[float],
    ): ...

class ParsedDatetimeRepr:
    year: Optional[int]
    month: Optional[int]
    day: Optional[int]
    hour: Optional[int]
    minute: Optional[int]
    second: Optional[int]

    def __init__(
        self,
        year: Optional[int],
        month: Optional[int],
        day: Optional[int],
        hour: Optional[int],
        minute: Optional[int],
        second: Optional[int],
    ): ...
